import copy
import math
from typing import Optional
import torch
from torch import nn
from torch.nn import functional as F

from module import commons
from module import modules
from module import attentions_onnx as attentions

from torch.nn import Conv1d, ConvTranspose1d, AvgPool1d, Conv2d
from torch.nn.utils import weight_norm, remove_weight_norm, spectral_norm
from module.commons import init_weights, get_padding
from module.quantize import ResidualVectorQuantizer
# from text import symbols
from text import symbols as symbols_v1
from text import symbols2 as symbols_v2
from torch.cuda.amp import autocast


class StochasticDurationPredictor(nn.Module):
    def __init__(
        self,
        in_channels,
        filter_channels,
        kernel_size,
        p_dropout,
        n_flows=4,
        gin_channels=0,
    ):
        super().__init__()
        filter_channels = in_channels  # it needs to be removed from future version.
        self.in_channels = in_channels
        self.filter_channels = filter_channels
        self.kernel_size = kernel_size
        self.p_dropout = p_dropout
        self.n_flows = n_flows
        self.gin_channels = gin_channels

        self.log_flow = modules.Log()
        self.flows = nn.ModuleList()
        self.flows.append(modules.ElementwiseAffine(2))
        for i in range(n_flows):
            self.flows.append(
                modules.ConvFlow(2, filter_channels, kernel_size, n_layers=3)
            )
            self.flows.append(modules.Flip())

        self.post_pre = nn.Conv1d(1, filter_channels, 1)
        self.post_proj = nn.Conv1d(filter_channels, filter_channels, 1)
        self.post_convs = modules.DDSConv(
            filter_channels, kernel_size, n_layers=3, p_dropout=p_dropout
        )
        self.post_flows = nn.ModuleList()
        self.post_flows.append(modules.ElementwiseAffine(2))
        for i in range(4):
            self.post_flows.append(
                modules.ConvFlow(2, filter_channels, kernel_size, n_layers=3)
            )
            self.post_flows.append(modules.Flip())

        self.pre = nn.Conv1d(in_channels, filter_channels, 1)
        self.proj = nn.Conv1d(filter_channels, filter_channels, 1)
        self.convs = modules.DDSConv(
            filter_channels, kernel_size, n_layers=3, p_dropout=p_dropout
        )
        if gin_channels != 0:
            self.cond = nn.Conv1d(gin_channels, filter_channels, 1)

    def forward(self, x, x_mask, w=None, g=None, reverse=False, noise_scale=1.0):
        x = torch.detach(x)
        x = self.pre(x)
        if g is not None:
            g = torch.detach(g)
            x = x + self.cond(g)
        x = self.convs(x, x_mask)
        x = self.proj(x) * x_mask

        if not reverse:
            flows = self.flows
            assert w is not None

            logdet_tot_q = 0
            h_w = self.post_pre(w)
            h_w = self.post_convs(h_w, x_mask)
            h_w = self.post_proj(h_w) * x_mask
            e_q = (
                torch.randn(w.size(0), 2, w.size(2)).to(device=x.device, dtype=x.dtype)
                * x_mask
            )
            z_q = e_q
            for flow in self.post_flows:
                z_q, logdet_q = flow(z_q, x_mask, g=(x + h_w))
                logdet_tot_q += logdet_q
            z_u, z1 = torch.split(z_q, [1, 1], 1)
            u = torch.sigmoid(z_u) * x_mask
            z0 = (w - u) * x_mask
            logdet_tot_q += torch.sum(
                (F.logsigmoid(z_u) + F.logsigmoid(-z_u)) * x_mask, [1, 2]
            )
            logq = (
                torch.sum(-0.5 * (math.log(2 * math.pi) + (e_q**2)) * x_mask, [1, 2])
                - logdet_tot_q
            )

            logdet_tot = 0
            z0, logdet = self.log_flow(z0, x_mask)
            logdet_tot += logdet
            z = torch.cat([z0, z1], 1)
            for flow in flows:
                z, logdet = flow(z, x_mask, g=x, reverse=reverse)
                logdet_tot = logdet_tot + logdet
            nll = (
                torch.sum(0.5 * (math.log(2 * math.pi) + (z**2)) * x_mask, [1, 2])
                - logdet_tot
            )
            return nll + logq  # [b]
        else:
            flows = list(reversed(self.flows))
            flows = flows[:-2] + [flows[-1]]  # remove a useless vflow
            z = (
                torch.randn(x.size(0), 2, x.size(2)).to(device=x.device, dtype=x.dtype)
                * noise_scale
            )
            for flow in flows:
                z = flow(z, x_mask, g=x, reverse=reverse)
            z0, z1 = torch.split(z, [1, 1], 1)
            logw = z0
            return logw


class DurationPredictor(nn.Module):
    def __init__(
        self, in_channels, filter_channels, kernel_size, p_dropout, gin_channels=0
    ):
        super().__init__()

        self.in_channels = in_channels
        self.filter_channels = filter_channels
        self.kernel_size = kernel_size
        self.p_dropout = p_dropout
        self.gin_channels = gin_channels

        self.drop = nn.Dropout(p_dropout)
        self.conv_1 = nn.Conv1d(
            in_channels, filter_channels, kernel_size, padding=kernel_size // 2
        )
        self.norm_1 = modules.LayerNorm(filter_channels)
        self.conv_2 = nn.Conv1d(
            filter_channels, filter_channels, kernel_size, padding=kernel_size // 2
        )
        self.norm_2 = modules.LayerNorm(filter_channels)
        self.proj = nn.Conv1d(filter_channels, 1, 1)

        if gin_channels != 0:
            self.cond = nn.Conv1d(gin_channels, in_channels, 1)

    def forward(self, x, x_mask, g=None):
        x = torch.detach(x)
        if g is not None:
            g = torch.detach(g)
            x = x + self.cond(g)
        x = self.conv_1(x * x_mask)
        x = torch.relu(x)
        x = self.norm_1(x)
        x = self.drop(x)
        x = self.conv_2(x * x_mask)
        x = torch.relu(x)
        x = self.norm_2(x)
        x = self.drop(x)
        x = self.proj(x * x_mask)
        return x * x_mask


class TextEncoder(nn.Module):
    def __init__(
        self,
        out_channels,
        hidden_channels,
        filter_channels,
        n_heads,
        n_layers,
        kernel_size,
        p_dropout,
        latent_channels=192,
        version="v2",
    ):
        super().__init__()
        self.out_channels = out_channels
        self.hidden_channels = hidden_channels
        self.filter_channels = filter_channels
        self.n_heads = n_heads
        self.n_layers = n_layers
        self.kernel_size = kernel_size
        self.p_dropout = p_dropout
        self.latent_channels = latent_channels
        self.version = version

        self.ssl_proj = nn.Conv1d(768, hidden_channels, 1)

        self.encoder_ssl = attentions.Encoder(
            hidden_channels,
            filter_channels,
            n_heads,
            n_layers // 2,
            kernel_size,
            p_dropout,
        )

        self.encoder_text = attentions.Encoder(
            hidden_channels, filter_channels, n_heads, n_layers, kernel_size, p_dropout
        )

        if self.version == "v1":
            symbols = symbols_v1.symbols
        else:
            symbols = symbols_v2.symbols
        self.text_embedding = nn.Embedding(len(symbols), hidden_channels)

        self.mrte = attentions.MRTE()

        self.encoder2 = attentions.Encoder(
            hidden_channels,
            filter_channels,
            n_heads,
            n_layers // 2,
            kernel_size,
            p_dropout,
        )

        self.proj = nn.Conv1d(hidden_channels, out_channels * 2, 1)

    def forward(self, y, text, ge, speed=1):
        y_mask = torch.ones_like(y[:1,:1,:])

        y = self.ssl_proj(y * y_mask) * y_mask
        y = self.encoder_ssl(y * y_mask, y_mask)

        text_mask = torch.ones_like(text).to(y.dtype).unsqueeze(0)

        text = self.text_embedding(text).transpose(1, 2)
        text = self.encoder_text(text * text_mask, text_mask)
        y = self.mrte(y, y_mask, text, text_mask, ge)

        y = self.encoder2(y * y_mask, y_mask)
        if(speed!=1):
            y = F.interpolate(y, size=int(y.shape[-1] / speed)+1, mode="linear")
            y_mask = F.interpolate(y_mask, size=y.shape[-1], mode="nearest")

        stats = self.proj(y) * y_mask
        m, logs = torch.split(stats, self.out_channels, dim=1)
        return y, m, logs, y_mask


class ResidualCouplingBlock(nn.Module):
    def __init__(
        self,
        channels,
        hidden_channels,
        kernel_size,
        dilation_rate,
        n_layers,
        n_flows=4,
        gin_channels=0,
    ):
        super().__init__()
        self.channels = channels
        self.hidden_channels = hidden_channels
        self.kernel_size = kernel_size
        self.dilation_rate = dilation_rate
        self.n_layers = n_layers
        self.n_flows = n_flows
        self.gin_channels = gin_channels

        self.flows = nn.ModuleList()
        for i in range(n_flows):
            self.flows.append(
                modules.ResidualCouplingLayer(
                    channels,
                    hidden_channels,
                    kernel_size,
                    dilation_rate,
                    n_layers,
                    gin_channels=gin_channels,
                    mean_only=True,
                )
            )
            self.flows.append(modules.Flip())

    def forward(self, x, x_mask, g=None, reverse=False):
        if not reverse:
            for flow in self.flows:
                x, _ = flow(x, x_mask, g=g, reverse=reverse)
        else:
            for flow in reversed(self.flows):
                x = flow(x, x_mask, g=g, reverse=reverse)
        return x


class PosteriorEncoder(nn.Module):
    def __init__(
        self,
        in_channels,
        out_channels,
        hidden_channels,
        kernel_size,
        dilation_rate,
        n_layers,
        gin_channels=0,
    ):
        super().__init__()
        self.in_channels = in_channels
        self.out_channels = out_channels
        self.hidden_channels = hidden_channels
        self.kernel_size = kernel_size
        self.dilation_rate = dilation_rate
        self.n_layers = n_layers
        self.gin_channels = gin_channels

        self.pre = nn.Conv1d(in_channels, hidden_channels, 1)
        self.enc = modules.WN(
            hidden_channels,
            kernel_size,
            dilation_rate,
            n_layers,
            gin_channels=gin_channels,
        )
        self.proj = nn.Conv1d(hidden_channels, out_channels * 2, 1)

    def forward(self, x, x_lengths, g=None):
        if g != None:
            g = g.detach()
        x_mask = torch.unsqueeze(commons.sequence_mask(x_lengths, x.size(2)), 1).to(
            x.dtype
        )
        x = self.pre(x) * x_mask
        x = self.enc(x, x_mask, g=g)
        stats = self.proj(x) * x_mask
        m, logs = torch.split(stats, self.out_channels, dim=1)
        z = (m + torch.randn_like(m) * torch.exp(logs)) * x_mask
        return z, m, logs, x_mask


class WNEncoder(nn.Module):
    def __init__(
        self,
        in_channels,
        out_channels,
        hidden_channels,
        kernel_size,
        dilation_rate,
        n_layers,
        gin_channels=0,
    ):
        super().__init__()
        self.in_channels = in_channels
        self.out_channels = out_channels
        self.hidden_channels = hidden_channels
        self.kernel_size = kernel_size
        self.dilation_rate = dilation_rate
        self.n_layers = n_layers
        self.gin_channels = gin_channels

        self.pre = nn.Conv1d(in_channels, hidden_channels, 1)
        self.enc = modules.WN(
            hidden_channels,
            kernel_size,
            dilation_rate,
            n_layers,
            gin_channels=gin_channels,
        )
        self.proj = nn.Conv1d(hidden_channels, out_channels, 1)
        self.norm = modules.LayerNorm(out_channels)

    def forward(self, x, x_lengths, g=None):
        x_mask = torch.unsqueeze(commons.sequence_mask(x_lengths, x.size(2)), 1).to(
            x.dtype
        )
        x = self.pre(x) * x_mask
        x = self.enc(x, x_mask, g=g)
        out = self.proj(x) * x_mask
        out = self.norm(out)
        return out


class Generator(torch.nn.Module):
    def __init__(
        self,
        initial_channel,
        resblock,
        resblock_kernel_sizes,
        resblock_dilation_sizes,
        upsample_rates,
        upsample_initial_channel,
        upsample_kernel_sizes,
        gin_channels=0,
    ):
        super(Generator, self).__init__()
        self.num_kernels = len(resblock_kernel_sizes)
        self.num_upsamples = len(upsample_rates)
        self.conv_pre = Conv1d(
            initial_channel, upsample_initial_channel, 7, 1, padding=3
        )
        resblock = modules.ResBlock1 if resblock == "1" else modules.ResBlock2

        self.ups = nn.ModuleList()
        for i, (u, k) in enumerate(zip(upsample_rates, upsample_kernel_sizes)):
            self.ups.append(
                weight_norm(
                    ConvTranspose1d(
                        upsample_initial_channel // (2**i),
                        upsample_initial_channel // (2 ** (i + 1)),
                        k,
                        u,
                        padding=(k - u) // 2,
                    )
                )
            )

        self.resblocks = nn.ModuleList()
        for i in range(len(self.ups)):
            ch = upsample_initial_channel // (2 ** (i + 1))
            for j, (k, d) in enumerate(
                zip(resblock_kernel_sizes, resblock_dilation_sizes)
            ):
                self.resblocks.append(resblock(ch, k, d))

        self.conv_post = Conv1d(ch, 1, 7, 1, padding=3, bias=False)
        self.ups.apply(init_weights)

        if gin_channels != 0:
            self.cond = nn.Conv1d(gin_channels, upsample_initial_channel, 1)

    def forward(self, x, g:Optional[torch.Tensor]=None):
        x = self.conv_pre(x)
        if g is not None:
            x = x + self.cond(g)

        for i in range(self.num_upsamples):
            x = F.leaky_relu(x, modules.LRELU_SLOPE)
            x = self.ups[i](x)
            xs = None
            for j in range(self.num_kernels):
                if xs is None:
                    xs = self.resblocks[i * self.num_kernels + j](x)
                else:
                    xs += self.resblocks[i * self.num_kernels + j](x)
            x = xs / self.num_kernels
        x = F.leaky_relu(x)
        x = self.conv_post(x)
        x = torch.tanh(x)

        return x

    def remove_weight_norm(self):
        print("Removing weight norm...")
        for l in self.ups:
            remove_weight_norm(l)
        for l in self.resblocks:
            l.remove_weight_norm()


class DiscriminatorP(torch.nn.Module):
    def __init__(self, period, kernel_size=5, stride=3, use_spectral_norm=False):
        super(DiscriminatorP, self).__init__()
        self.period = period
        self.use_spectral_norm = use_spectral_norm
        norm_f = weight_norm if use_spectral_norm == False else spectral_norm
        self.convs = nn.ModuleList(
            [
                norm_f(
                    Conv2d(
                        1,
                        32,
                        (kernel_size, 1),
                        (stride, 1),
                        padding=(get_padding(kernel_size, 1), 0),
                    )
                ),
                norm_f(
                    Conv2d(
                        32,
                        128,
                        (kernel_size, 1),
                        (stride, 1),
                        padding=(get_padding(kernel_size, 1), 0),
                    )
                ),
                norm_f(
                    Conv2d(
                        128,
                        512,
                        (kernel_size, 1),
                        (stride, 1),
                        padding=(get_padding(kernel_size, 1), 0),
                    )
                ),
                norm_f(
                    Conv2d(
                        512,
                        1024,
                        (kernel_size, 1),
                        (stride, 1),
                        padding=(get_padding(kernel_size, 1), 0),
                    )
                ),
                norm_f(
                    Conv2d(
                        1024,
                        1024,
                        (kernel_size, 1),
                        1,
                        padding=(get_padding(kernel_size, 1), 0),
                    )
                ),
            ]
        )
        self.conv_post = norm_f(Conv2d(1024, 1, (3, 1), 1, padding=(1, 0)))

    def forward(self, x):
        fmap = []

        # 1d to 2d
        b, c, t = x.shape
        if t % self.period != 0:  # pad first
            n_pad = self.period - (t % self.period)
            x = F.pad(x, (0, n_pad), "reflect")
            t = t + n_pad
        x = x.view(b, c, t // self.period, self.period)

        for l in self.convs:
            x = l(x)
            x = F.leaky_relu(x, modules.LRELU_SLOPE)
            fmap.append(x)
        x = self.conv_post(x)
        fmap.append(x)
        x = torch.flatten(x, 1, -1)

        return x, fmap


class DiscriminatorS(torch.nn.Module):
    def __init__(self, use_spectral_norm=False):
        super(DiscriminatorS, self).__init__()
        norm_f = weight_norm if use_spectral_norm == False else spectral_norm
        self.convs = nn.ModuleList(
            [
                norm_f(Conv1d(1, 16, 15, 1, padding=7)),
                norm_f(Conv1d(16, 64, 41, 4, groups=4, padding=20)),
                norm_f(Conv1d(64, 256, 41, 4, groups=16, padding=20)),
                norm_f(Conv1d(256, 1024, 41, 4, groups=64, padding=20)),
                norm_f(Conv1d(1024, 1024, 41, 4, groups=256, padding=20)),
                norm_f(Conv1d(1024, 1024, 5, 1, padding=2)),
            ]
        )
        self.conv_post = norm_f(Conv1d(1024, 1, 3, 1, padding=1))

    def forward(self, x):
        fmap = []

        for l in self.convs:
            x = l(x)
            x = F.leaky_relu(x, modules.LRELU_SLOPE)
            fmap.append(x)
        x = self.conv_post(x)
        fmap.append(x)
        x = torch.flatten(x, 1, -1)

        return x, fmap


class MultiPeriodDiscriminator(torch.nn.Module):
    def __init__(self, use_spectral_norm=False):
        super(MultiPeriodDiscriminator, self).__init__()
        periods = [2, 3, 5, 7, 11]

        discs = [DiscriminatorS(use_spectral_norm=use_spectral_norm)]
        discs = discs + [
            DiscriminatorP(i, use_spectral_norm=use_spectral_norm) for i in periods
        ]
        self.discriminators = nn.ModuleList(discs)

    def forward(self, y, y_hat):
        y_d_rs = []
        y_d_gs = []
        fmap_rs = []
        fmap_gs = []
        for i, d in enumerate(self.discriminators):
            y_d_r, fmap_r = d(y)
            y_d_g, fmap_g = d(y_hat)
            y_d_rs.append(y_d_r)
            y_d_gs.append(y_d_g)
            fmap_rs.append(fmap_r)
            fmap_gs.append(fmap_g)

        return y_d_rs, y_d_gs, fmap_rs, fmap_gs


class ReferenceEncoder(nn.Module):
    """
    inputs --- [N, Ty/r, n_mels*r]  mels
    outputs --- [N, ref_enc_gru_size]
    """

    def __init__(self, spec_channels, gin_channels=0):
        super().__init__()
        self.spec_channels = spec_channels
        ref_enc_filters = [32, 32, 64, 64, 128, 128]
        K = len(ref_enc_filters)
        filters = [1] + ref_enc_filters
        convs = [
            weight_norm(
                nn.Conv2d(
                    in_channels=filters[i],
                    out_channels=filters[i + 1],
                    kernel_size=(3, 3),
                    stride=(2, 2),
                    padding=(1, 1),
                )
            )
            for i in range(K)
        ]
        self.convs = nn.ModuleList(convs)
        # self.wns = nn.ModuleList([weight_norm(num_features=ref_enc_filters[i]) for i in range(K)])

        out_channels = self.calculate_channels(spec_channels, 3, 2, 1, K)
        self.gru = nn.GRU(
            input_size=ref_enc_filters[-1] * out_channels,
            hidden_size=256 // 2,
            batch_first=True,
        )
        self.proj = nn.Linear(128, gin_channels)

    def forward(self, inputs):
        N = inputs.size(0)
        out = inputs.view(N, 1, -1, self.spec_channels)  # [N, 1, Ty, n_freqs]
        for conv in self.convs:
            out = conv(out)
            # out = wn(out)
            out = F.relu(out)  # [N, 128, Ty//2^K, n_mels//2^K]

        out = out.transpose(1, 2)  # [N, Ty//2^K, 128, n_mels//2^K]
        T = out.size(1)
        N = out.size(0)
        out = out.contiguous().view(N, T, -1)  # [N, Ty//2^K, 128*n_mels//2^K]

        self.gru.flatten_parameters()
        memory, out = self.gru(out)  # out --- [1, N, 128]

        return self.proj(out.squeeze(0)).unsqueeze(-1)

    def calculate_channels(self, L, kernel_size, stride, pad, n_convs):
        for i in range(n_convs):
            L = (L - kernel_size + 2 * pad) // stride + 1
        return L


class Quantizer_module(torch.nn.Module):
    def __init__(self, n_e, e_dim):
        super(Quantizer_module, self).__init__()
        self.embedding = nn.Embedding(n_e, e_dim)
        self.embedding.weight.data.uniform_(-1.0 / n_e, 1.0 / n_e)

    def forward(self, x):
        d = (
            torch.sum(x**2, 1, keepdim=True)
            + torch.sum(self.embedding.weight**2, 1)
            - 2 * torch.matmul(x, self.embedding.weight.T)
        )
        min_indicies = torch.argmin(d, 1)
        z_q = self.embedding(min_indicies)
        return z_q, min_indicies


class Quantizer(torch.nn.Module):
    def __init__(self, embed_dim=512, n_code_groups=4, n_codes=160):
        super(Quantizer, self).__init__()
        assert embed_dim % n_code_groups == 0
        self.quantizer_modules = nn.ModuleList(
            [
                Quantizer_module(n_codes, embed_dim // n_code_groups)
                for _ in range(n_code_groups)
            ]
        )
        self.n_code_groups = n_code_groups
        self.embed_dim = embed_dim

    def forward(self, xin):
        # B, C, T
        B, C, T = xin.shape
        xin = xin.transpose(1, 2)
        x = xin.reshape(-1, self.embed_dim)
        x = torch.split(x, self.embed_dim // self.n_code_groups, dim=-1)
        min_indicies = []
        z_q = []
        for _x, m in zip(x, self.quantizer_modules):
            _z_q, _min_indicies = m(_x)
            z_q.append(_z_q)
            min_indicies.append(_min_indicies)  # B * T,
        z_q = torch.cat(z_q, -1).reshape(xin.shape)
        loss = 0.25 * torch.mean((z_q.detach() - xin) ** 2) + torch.mean(
            (z_q - xin.detach()) ** 2
        )
        z_q = xin + (z_q - xin).detach()
        z_q = z_q.transpose(1, 2)
        codes = torch.stack(min_indicies, -1).reshape(B, T, self.n_code_groups)
        return z_q, loss, codes.transpose(1, 2)

    def embed(self, x):
        # idx: N, 4, T
        x = x.transpose(1, 2)
        x = torch.split(x, 1, 2)
        ret = []
        for q, embed in zip(x, self.quantizer_modules):
            q = embed.embedding(q.squeeze(-1))
            ret.append(q)
        ret = torch.cat(ret, -1)
        return ret.transpose(1, 2)  # N, C, T


class CodePredictor(nn.Module):
    def __init__(
        self,
        hidden_channels,
        filter_channels,
        n_heads,
        n_layers,
        kernel_size,
        p_dropout,
        n_q=8,
        dims=1024,
        ssl_dim=768,
    ):
        super().__init__()
        self.hidden_channels = hidden_channels
        self.filter_channels = filter_channels
        self.n_heads = n_heads
        self.n_layers = n_layers
        self.kernel_size = kernel_size
        self.p_dropout = p_dropout

        self.vq_proj = nn.Conv1d(ssl_dim, hidden_channels, 1)
        self.ref_enc = modules.MelStyleEncoder(
            ssl_dim, style_vector_dim=hidden_channels
        )

        self.encoder = attentions.Encoder(
            hidden_channels, filter_channels, n_heads, n_layers, kernel_size, p_dropout
        )

        self.out_proj = nn.Conv1d(hidden_channels, (n_q - 1) * dims, 1)
        self.n_q = n_q
        self.dims = dims

    def forward(self, x, x_mask, refer, codes, infer=False):
        x = x.detach()
        x = self.vq_proj(x * x_mask) * x_mask
        g = self.ref_enc(refer, x_mask)
        x = x + g
        x = self.encoder(x * x_mask, x_mask)
        x = self.out_proj(x * x_mask) * x_mask
        logits = x.reshape(x.shape[0], self.n_q - 1, self.dims, x.shape[-1]).transpose(
            2, 3
        )
        target = codes[1:].transpose(0, 1)
        if not infer:
            logits = logits.reshape(-1, self.dims)
            target = target.reshape(-1)
            loss = torch.nn.functional.cross_entropy(logits, target)
            return loss
        else:
            _, top10_preds = torch.topk(logits, 10, dim=-1)
            correct_top10 = torch.any(top10_preds == target.unsqueeze(-1), dim=-1)
            top3_acc = 100 * torch.mean(correct_top10.float()).detach().cpu().item()

            print("Top-10 Accuracy:", top3_acc, "%")

            pred_codes = torch.argmax(logits, dim=-1)
            acc = 100 * torch.mean((pred_codes == target).float()).detach().cpu().item()
            print("Top-1 Accuracy:", acc, "%")

            return pred_codes.transpose(0, 1)


class SynthesizerTrn(nn.Module):
    """
    Synthesizer for Training
    """

    def __init__(
        self,
        spec_channels,
        segment_size,
        inter_channels,
        hidden_channels,
        filter_channels,
        n_heads,
        n_layers,
        kernel_size,
        p_dropout,
        resblock,
        resblock_kernel_sizes,
        resblock_dilation_sizes,
        upsample_rates,
        upsample_initial_channel,
        upsample_kernel_sizes,
        n_speakers=0,
        gin_channels=0,
        use_sdp=True,
        semantic_frame_rate=None,
        freeze_quantizer=None,
        version="v2",
        **kwargs
    ):
        super().__init__()
        self.spec_channels = spec_channels
        self.inter_channels = inter_channels
        self.hidden_channels = hidden_channels
        self.filter_channels = filter_channels
        self.n_heads = n_heads
        self.n_layers = n_layers
        self.kernel_size = kernel_size
        self.p_dropout = p_dropout
        self.resblock = resblock
        self.resblock_kernel_sizes = resblock_kernel_sizes
        self.resblock_dilation_sizes = resblock_dilation_sizes
        self.upsample_rates = upsample_rates
        self.upsample_initial_channel = upsample_initial_channel
        self.upsample_kernel_sizes = upsample_kernel_sizes
        self.segment_size = segment_size
        self.n_speakers = n_speakers
        self.gin_channels = gin_channels
        self.version = version

        self.use_sdp = use_sdp
        self.enc_p = TextEncoder(
            inter_channels,
            hidden_channels,
            filter_channels,
            n_heads,
            n_layers,
            kernel_size,
            p_dropout,
            version=version,
        )
        self.dec = Generator(
            inter_channels,
            resblock,
            resblock_kernel_sizes,
            resblock_dilation_sizes,
            upsample_rates,
            upsample_initial_channel,
            upsample_kernel_sizes,
            gin_channels=gin_channels,
        )
        # self.enc_q = PosteriorEncoder(
        #     spec_channels,
        #     inter_channels,
        #     hidden_channels,
        #     5,
        #     1,
        #     16,
        #     gin_channels=gin_channels,
        # )
        self.flow = ResidualCouplingBlock(
            inter_channels, hidden_channels, 5, 1, 4, gin_channels=gin_channels
        )

        # self.version=os.environ.get("version","v1")
        if self.version == "v1":
            self.ref_enc = modules.MelStyleEncoder(spec_channels, style_vector_dim=gin_channels)
        else:
            self.ref_enc = modules.MelStyleEncoder(704, style_vector_dim=gin_channels)

        ssl_dim = 768
        self.ssl_dim = ssl_dim
        assert semantic_frame_rate in ["25hz", "50hz"]
        self.semantic_frame_rate = semantic_frame_rate
        if semantic_frame_rate == "25hz":
            self.ssl_proj = nn.Conv1d(ssl_dim, ssl_dim, 2, stride=2)
        else:
            self.ssl_proj = nn.Conv1d(ssl_dim, ssl_dim, 1, stride=1)

        self.quantizer = ResidualVectorQuantizer(dimension=ssl_dim, n_q=1, bins=1024)
        if freeze_quantizer:
            self.ssl_proj.requires_grad_(False)
            self.quantizer.requires_grad_(False)
            # self.enc_p.text_embedding.requires_grad_(False)
            # self.enc_p.encoder_text.requires_grad_(False)
            # self.enc_p.mrte.requires_grad_(False)

    def forward(self, codes, text, refer,noise_scale=0.5, speed=1):
        refer_mask = torch.ones_like(refer[:1,:1,:])
        if (self.version == "v1"):
            ge = self.ref_enc(refer * refer_mask, refer_mask)
        else:
            ge = self.ref_enc(refer[:, :704] * refer_mask, refer_mask)

        quantized = self.quantizer.decode(codes)
        if self.semantic_frame_rate == "25hz":
            dquantized = torch.cat([quantized, quantized]).permute(1, 2, 0)
            quantized = dquantized.contiguous().view(1, self.ssl_dim, -1)

        x, m_p, logs_p, y_mask = self.enc_p(
            quantized, text, ge, speed
        )
        
        z_p = m_p + torch.randn_like(m_p) * torch.exp(logs_p) * noise_scale

        z = self.flow(z_p, y_mask, g=ge, reverse=True)

        o = self.dec((z * y_mask)[:, :, :], g=ge)
        return o

    def extract_latent(self, x):
        ssl = self.ssl_proj(x)
        quantized, codes, commit_loss, quantized_list = self.quantizer(ssl)
        return codes.transpose(0, 1)


from ..data_classes.realtime_server_proto import (
    InputAudioBufferSpeechStarted,
    InputAudioBufferSpeechStopped,
    InputAudioBufferCommitted,
    InputAudioBufferCleared,
    MessageContentServer,
    ConversationItemMessageServer,
    ContentTypeServer
)
# from ..logger.logger import logger
from gallama.realtime.websocket_client import WebSocketClient
from gallama.realtime.websocket_session import WebSocketSession
from fastapi import WebSocket
from .vad import VADProcessor
import base64
from gallama.data_classes.realtime_client_proto import *
from gallama.data_classes.internal_ws import *
from ..dependencies_server import get_server_logger
import numpy as np
import time
import asyncio

logger = get_server_logger()

class WebSocketMessageHandler:
    """
    Handles different message types
    This class is the entry point for all client messages.
    This class should refrain from actually doing any processing, but pass on the job to the appropriate handler.
    """
    def __init__(self):
        pass

    def initialize(self, stt_url: str, llm_url: str, tts_url: str):
        self.stt_url = stt_url
        self.llm_url = llm_url
        self.tts_url = tts_url

        self.ws_stt = WebSocketClient(stt_url)
        self.ws_llm = WebSocketClient(llm_url)
        self.ws_tts = WebSocketClient(tts_url)

        self.initialize_ws = False
        self.buffer_lock = asyncio.Lock()

        self.handlers = {
            # session
            "session.update": self._session_update,

            # audio
            "input_audio_buffer.append": self._input_audio_buffer_append,
            "input_audio_buffer.commit": self._input_audio_buffer_commit,
            "input_audio_buffer.clear": self._input_audio_buffer_clear,

            # converssation
            "conversation.item.create": self._conversation_item_create,
            "conversation.item.truncate": self._conversation_item_truncate,
            "conversation.item.delete": self._conversation_item_delete,

            # response
            "response.create": self._response_create,
            "response.cancel": self._response_cancel,

        }

    async def initialize_connection(self):
        await self.ws_stt.connect()

        await self.ws_llm.connect()
        await self.ws_tts.connect()
        self.initialize_ws = True

    async def cleanup(self):
        # send signal to respective web socket to clean data
        await self.ws_llm.send_pydantic_message(WSInterCleanup())
        await self.ws_tts.send_pydantic_message(WSInterCleanup())
        await self.ws_stt.send_pydantic_message(WSInterCleanup())



    async def handle_message(self, websocket: WebSocket, session: WebSocketSession, message: dict):
        """This is the entrypoint for client messages. It dispatches the message to the appropriate handler."""
        handler = self.handlers.get(message["type"])
        if handler:
            await handler(websocket, session, message)
        else:
            logger.warning(f"Unknown message type: {message['type']}")

    async def _session_update(self, websocket: WebSocket, session: WebSocketSession, message: dict):
        update_config = SessionConfig(**message["session"])

        if 'voice' in message["session"] and session.voice_used:
            del message["session"]["voice"]

        session.config = SessionConfig(**{**session.config.model_dump(), **message["session"]})

        # send respective ws update
        await self.ws_llm.send_pydantic_message(WSInterConfigUpdate(config=session.config))
        await self.ws_tts.send_pydantic_message(WSInterConfigUpdate(config=session.config))
        await self.ws_stt.send_pydantic_message(WSInterConfigUpdate(config=session.config))

        logger.info(f"Session updated: {session.config.model_dump()}")


        await websocket.send_json({
            "event_id": await session.queues.next_event(),
            "type": "session.updated",
            "session": {
                "id": session.id,
                "object": "realtime.session",
                "model": "default model",
                **session.config.model_dump()
            }
        })

    async def _input_audio_buffer_append(self, websocket: WebSocket, session: WebSocketSession, message: dict):
        try:
            # Decode and convert audio once
            audio_bytes = base64.b64decode(message["audio"])
            audio_data = np.frombuffer(audio_bytes, dtype=np.int16)
            audio_float = audio_data.astype(np.float32) / 32768.0
            await session.queues.append_unprocessed_audio(
                message["audio"],
                ws_stt=self.ws_stt,
                audio_float=audio_float
            )

        except Exception as e:
            logger.error(f"Error processing audio: {str(e)}")
            raise

    async def _input_audio_buffer_commit(self, websocket: WebSocket, session: WebSocketSession, message: dict, item_id: str = None):
        # if is none, get a new item_id
        if item_id is None:
            item_id = await session.queues.next_item()

        # handling internal audio commit
        await session.queues.commit_unprocessed_audio(ws_stt=self.ws_stt, item_id=item_id)

        #   send user committed msg
        await websocket.send_json(InputAudioBufferCommitted(**{
            "event_id": await session.queues.next_event(),
            "type": "input_audio_buffer.committed",
            "previous_item_id": await session.queues.current_item_id(),
            "item_id": item_id
        }).model_dump())


    async def _input_audio_buffer_clear(self, websocket: WebSocket, session: WebSocketSession, message: dict):
        # handling internal audio buffer clear
        await session.queues.clear_unprocessed_audio(ws_stt=self.ws_stt)

        #   send user committed msg
        await websocket.send_json(InputAudioBufferCleared(**{
            "event_id": await session.queues.next_event(),
            "type": "input_audio_buffer.cleared",
        }).model_dump())


    async def _conversation_item_create(self, websocket: WebSocket, session: WebSocketSession, message: dict):
        # for conversation.item.created, it is a completed item, hence no need to do streaming
        event = ConversationItemCreate(**message)
        await session.queues.unprocessed.put(event)


    async def _conversation_item_truncate(self, websocket: WebSocket, session: WebSocketSession, message: dict):
        event = ConversationItemTruncate(**message)
        await session.queues.unprocessed.put(event)

    async def _conversation_item_delete(self, websocket: WebSocket, session: WebSocketSession, message: dict):
        event = ConversationItemDelete(**message)
        await session.queues.unprocessed.put(event.item)


    async def _response_create(self, websocket: WebSocket, session: WebSocketSession, message: dict):
        # add a Conversation Item into unprocess queue
        # once the processing reach this model, it will trigger llm -> tts

        if message.get("response"):
            # combine the config with existing
            # else it will use the existing
            if message["response"].get("conversation"):
                message["response"].pop("conversation")     # currently not supported

            if message["response"]:     # if there are other field, combine with existing setting
                message["response"] = {
                    **session.config.model_dump(),
                    **message["response"]
                }
            elif message["response"] == {}:
                message.pop("response")


        item = ResponseCreate(**message)
        await session.queues.unprocessed.put(item)

    async def _response_cancel(self, websocket: WebSocket, session: WebSocketSession, message: dict):
        """Handle response.cancel event"""
        try:
            async with session.current_response_lock:
                if not session.current_response:
                    logger.warning("No current response to cancel")
                else:
                    # Since response is handled synchronously in process_unprocessed_queue,
                    # we just need to add a cancel event to the queue
                    cancel_event = ResponseCancel(**message)
                    await session.current_response.cancel(cancel_event)

        except Exception as e:
            logger.error(f"Error in response cancellation: {str(e)}")
            await websocket.send_json({
                "type": "error",
                "error": {
                    "code": "internal_error",
                    "message": "Failed to cancel response"
                }
            })




import librosa
import numpy as np
import io
import soundfile as sf
from gallama.logger.logger import logger
import time
import asyncio
from ..data_classes import TTSEvent


class AudioProcessor:
    def __init__(
        self,
        input_sample_rate: int = 32000,
        output_sample_rate: int = 24000,
        min_chunk_size: int = 1024,
        optimal_chunk_size: int = 32768,
        format='wav',
        subtype='PCM_16'
    ):
        self.input_sample_rate = input_sample_rate
        self.output_sample_rate = output_sample_rate
        self.format = format
        self.subtype = subtype
        self.buffer = bytearray()
        self.min_chunk_size = min_chunk_size
        self.optimal_chunk_size = optimal_chunk_size
        self.total_processed = 0
        self.chunks_processed = 0
        self.process_start_time = None
        self.bytes_per_sample = 2  # For PCM16

    def process_chunk(self, audio_data, sample_rate=None):
        """Process audio chunk with dynamic chunk sizing and buffer management"""
        MAX_BUFFER_SIZE = 1024 * 1024  # 1MB maximum buffer size

        if self.process_start_time is None:
            self.process_start_time = time.time()

        if sample_rate is not None:
            self.input_sample_rate = sample_rate

        # Convert numpy array to bytes if needed
        if isinstance(audio_data, np.ndarray):
            audio_data = audio_data.tobytes()

        original_buffer_size = len(self.buffer)
        self.buffer.extend(audio_data)
        new_buffer_size = len(self.buffer)

        logger.info(
            f"Buffer state - Previous: {original_buffer_size} bytes, Added: {len(audio_data)} bytes, New: {new_buffer_size} bytes")

        # Emergency buffer size control
        if new_buffer_size > MAX_BUFFER_SIZE:
            logger.warning(f"Buffer exceeded {MAX_BUFFER_SIZE} bytes, processing aggressively")
            results = []
            # Process in larger chunks until buffer is under control
            while len(self.buffer) > MAX_BUFFER_SIZE // 2:
                chunk_size = min(len(self.buffer) // 2, self.optimal_chunk_size * 4)
                chunk = self.buffer[:chunk_size * self.bytes_per_sample]
                result = self._process_audio_chunk(chunk)
                if result:
                    results.append(result)
                self.buffer = self.buffer[chunk_size * self.bytes_per_sample:]
            return b''.join(results) if results else None

        # Check if we have enough data
        samples_available = len(self.buffer) // self.bytes_per_sample
        logger.info(f"Samples available: {samples_available}, Minimum needed: {self.min_chunk_size}")

        if samples_available < self.min_chunk_size:
            logger.info("Not enough samples, buffering...")
            return None

        # Dynamic chunk sizing based on buffer fullness
        target_chunk_size = self.optimal_chunk_size
        if samples_available > self.optimal_chunk_size * 2:
            # Gradually increase chunk size as buffer grows
            buffer_multiple = samples_available / self.optimal_chunk_size
            target_chunk_size = min(
                int(self.optimal_chunk_size * min(buffer_multiple / 2, 4)),
                samples_available // 2
            )
            logger.info(f"Buffer getting full, processing larger chunk: {target_chunk_size} samples")

        # Extract chunk to process
        bytes_to_process = target_chunk_size * self.bytes_per_sample
        chunk = self.buffer[:bytes_to_process]
        self.buffer = self.buffer[bytes_to_process:]

        self.chunks_processed += 1
        self.total_processed += target_chunk_size

        processing_time = time.time() - self.process_start_time
        logger.info(
            f"Processed {self.chunks_processed} chunks, {self.total_processed} samples in {processing_time:.2f}s")
        logger.info(f"Average throughput: {self.total_processed / processing_time:.2f} samples/second")

        return self._process_audio_chunk(chunk)

    def _process_audio_chunk(self, audio_bytes):
        """Process a single chunk of audio data with optimized resampling"""
        BATCH_SIZE = 32768  # 32KB batch size for large chunks
        process_start = time.time()

        try:
            # Convert bytes to numpy array
            audio_array = np.frombuffer(audio_bytes, dtype=np.int16)

            # Only resample if rates differ
            if self.input_sample_rate != self.output_sample_rate:
                # Convert to float32 for librosa
                audio_float = audio_array.astype(np.float32) / 32768.0

                # Use faster resampling algorithm for real-time processing
                audio_resampled = librosa.resample(
                    audio_float,
                    orig_sr=self.input_sample_rate,
                    target_sr=self.output_sample_rate,
                    res_type='kaiser_fast'  # Faster resampling algorithm
                )

                # Convert back to int16
                audio_array = (audio_resampled * 32768).astype(np.int16)

            # For large chunks, process all the data at once instead of batching
            buffer = io.BytesIO()
            sf.write(
                buffer,
                audio_array,
                self.output_sample_rate,
                format=self.format,
                subtype=self.subtype
            )
            buffer.seek(0)

            process_duration = time.time() - process_start
            if process_duration > 0.1:  # Log warning if processing takes too long
                logger.warning(f"Chunk processing took {process_duration * 1000:.2f}ms")

            return buffer.getvalue()

        except Exception as e:
            logger.error(f"Error in audio chunk processing: {str(e)}")
            return None


    def flush(self):
        """Process any remaining audio in the buffer"""
        if len(self.buffer) >= self.min_chunk_size * self.bytes_per_sample:
            logger.info(f"Flushing remaining {len(self.buffer)} bytes from buffer")
            result = self._process_audio_chunk(self.buffer)
            self.buffer = bytearray()
            return result
        return None


from fastapi import FastAPI, HTTPException, Request
import json
import textwrap
from copy import deepcopy
from gallama.data_classes import AgentWithThinking, MixtureOfAgents
from typing import List, Dict, Union
from fastapi.responses import Response, StreamingResponse
from .request_handler import forward_request
import asyncio


async def modify_request(request: Request, changes: Dict[str, any]):
    body = await request.json()
    body_copy = deepcopy(body)

    # Apply the changes to the body
    for key, value in changes.items():
        if key in body_copy:
            body_copy[key] = value

    modified_body = json.dumps(body_copy).encode()

    # Create a copy of the headers and update the Content-Length
    modified_headers = dict(request.headers)
    modified_headers["content-length"] = str(len(modified_body))

    return modified_body, modified_headers


async def consolidate_responses(original_request: Request, responses: List[Response]):
    # This is a placeholder implementation. Replace with your actual consolidation logic.
    consolidation_prompt = "\n---\n" + textwrap.dedent("""
    Please synthesize the following reference answer into a single, high-quality response.
    It is crucial to critically evaluate the information provided in these responses, recognizing that some of it may be biased or incorrect.
    """).strip() + "\n---\n"

    for index, response in enumerate(responses):
        response_body = json.loads(response.body)
        response_msg = response_body["choices"][0].get("message")

        # add the answer to the prompt
        consolidation_prompt += f"Reference answer {str(index+1)}\n"
        if response_msg and response_msg.get("content"):
            consolidation_prompt += response_msg["content"] + "\n"

        if response_msg and response_msg.get("tool_calls"):
            for tool_call in response_msg.get("tool_calls"):
                consolidation_prompt += str(tool_call["function"]) + "\n"

        consolidation_prompt += "---\n"

    consolidation_prompt += "Now provide the final synthesized response:\n"

    # Create a modified version of request with the answers from respective agents
    original_body = await original_request.json()

    modified_messages = original_body.get("messages", [])

    # append the final synthesis prompt to the message list
    modified_messages.append({
        "role": "user",
        "content": consolidation_prompt
    })

    modified_body, modified_headers = await modify_request(
        original_request,
        changes={"messages": modified_messages}
    )

    return modified_body, modified_headers


async def forward_to_multiple_agents(request: Request, agent_list: List[Union[str, AgentWithThinking]], modified_body: str, modified_headers: str, models: Dict, active_requests: Dict):
    tasks = []
    for agent in agent_list:
        if isinstance(agent, str):
            instance = await get_instance_for_model(agent, models, active_requests)
            tasks.append(forward_request(request, instance, modified_body, modified_headers))
        elif isinstance(agent, AgentWithThinking):
            instance = await get_instance_for_model(agent.model, models, active_requests)
            if agent.thinking_template:
                # Modify the request to include the thinking in the thinking_template field
                agent_body = json.loads(modified_body)
                agent_body["thinking_template"] = agent.thinking_template
                agent_modified_body = json.dumps(agent_body).encode()
                agent_modified_headers = dict(modified_headers)
                agent_modified_headers["content-length"] = str(len(agent_modified_body))
                tasks.append(forward_request(request, instance, agent_modified_body, agent_modified_headers))
            else:
                tasks.append(forward_request(request, instance, modified_body, modified_headers))
    return await asyncio.gather(*tasks)


async def get_instance_for_model(model: str, models: Dict, active_requests: Dict):
    running_instances = [inst for inst in models[model].instances if inst.status == "running"]
    if not running_instances:
        raise HTTPException(status_code=503, detail=f"No running instances available for model: {model}")

    return min(running_instances, key=lambda inst: active_requests[inst.port])


async def handle_mixture_of_agent_request(request: Request, body_json: dict, models: Dict, active_requests: Dict):
    moa_config = MixtureOfAgents(**body_json.get("mixture_of_agents"))
    agent_list = moa_config.agent_list
    master_agent = body_json.get("model")

    if not agent_list or not master_agent:
        raise HTTPException(status_code=400, detail="Invalid request: missing agent_list or master_agent")

    # Validate all models upfront
    all_models = [agent.model if isinstance(agent, AgentWithThinking) else agent for agent in agent_list] + [master_agent]

    # Create a modified request where streaming is turned off
    changes = {"stream": False}
    modified_body, modified_headers = await modify_request(request, changes)

    # Forward the request to all models in the agent_list
    responses = await forward_to_multiple_agents(request, agent_list, modified_body, modified_headers, models, active_requests)

    # Consolidate responses
    consol_modified_body, consol_modified_headers = await consolidate_responses(request, responses)

    # Forward consolidated response to master_agent
    instance = await get_instance_for_model(master_agent, models, active_requests)
    final_response = await forward_request(request, instance, consol_modified_body, consol_modified_headers)
    return final_response




from ..base import ASRBase
from typing import Literal, List, Union, BinaryIO, Optional
import numpy as np

from ....logger import logger
from ....data_classes import ModelSpec, TranscriptionResponse, TimeStampedWord, LanguageType
import soundfile as sf


# faster whisper wont work. For now we import to reuse the data classes
# TODO create the data classes in gallama
try:
    from faster_whisper.transcribe import TranscriptionOptions, TranscriptionInfo, Segment
except ImportError:
    TranscriptionOptions, TranscriptionInfo, Segment = None, None, None, None, None


import mlx_whisper as mlx_model
import mlx.core as mx


class ASRMLXWhisper(ASRBase):
    """
    An implementation of ASRBase that uses the MLX library as backend.
    """


    def load_model(self, model_spec: ModelSpec):
        """
        For MLX, we simply store the model parameters for use with the MLX transcribe function.
        """
        self.model_id = model_spec.model_id  # path or HF repo for the MLX converted weights
        self.model_name = model_spec.model_name
        self.quant = model_spec.quant

        # mlx whisper doesnt have dedicated load method
        # but it will load when transcribe is ran
        mlx_model.transcribe(
            audio=np.random.randn(44100 * 2),
            path_or_hf_repo=model_spec.model_id,
        )

        # Use fp16 if quantization is "16.0"; otherwise, assume fp32.
        # self.fp16 = True if self.quant == "16.0" else False
        return mlx_model

    # @staticmethod
    # def binaryio_to_numpy(audio_io: BinaryIO) -> Tuple[np.ndarray, int]:
    #     """
    #     Convert a BinaryIO audio stream into a numpy array.
    #     Returns a tuple of (audio_data, sample_rate).
    #     """
    #     audio_io.seek(0)  # Ensure the file pointer is at the beginning
    #     audio_data, sample_rate = sf.read(audio_io)
    #     return audio_data, sample_rate

    def transcribe_to_segment(
        self,
        audio: Union[str, np.ndarray],
        init_prompt: str = "",
        temperature: float = 0.0,
        language: LanguageType = None,
        batch: bool = False,
        batch_size: int = 8,
    ) -> List[dict]:
        """
        Transcribe audio into a list of segments.
        Calls the MLX transcribe function with word timestamps enabled.
        """

        if not isinstance(audio, np.ndarray):
            audio_to_use, sample_rate = sf.read(audio)
        else:
            audio_to_use = audio

        response = self.model.transcribe(
            audio=audio_to_use,
            path_or_hf_repo=self.model_id,  # eventhough loaded, we still need to use this
            temperature=temperature,
            condition_on_previous_text=True,
            initial_prompt=init_prompt,
            word_timestamps=True,
            hallucination_silence_threshold=1,
            **self.transcribe_kargs  # any extra keyword arguments set via use_vad(), etc.
        )
        segments = response.get("segments", [])
        return segments

    def transcribe(
        self,
        audio: Union[str, np.ndarray],
        init_prompt: str = "",
        temperature: float = 0.0,
        language=None,
        include_segments: bool = False,
        batch: bool = False,
    ) -> TranscriptionResponse:
        """
        Transcribe audio to full text and return a TranscriptionResponse.
        The response includes the text, detected language, duration (from the last segment),
        and word-level timestamps (either from MLX if available, or approximated).
        """

        if not isinstance(audio, np.ndarray):
            audio_to_use, sample_rate = sf.read(audio)
        else:
            audio_to_use = audio


        response = self.model.transcribe(
            audio=audio_to_use,
            path_or_hf_repo=self.model_id,  # eventhough loaded, we still need to use this
            temperature=temperature,
            condition_on_previous_text=True,
            initial_prompt=init_prompt,
            word_timestamps=True,
            hallucination_silence_threshold=1,
            **self.transcribe_kargs  # any extra keyword arguments set via use_vad(), etc.
        )
        text = response.get("text", "")
        segments = response.get("segments", [])
        lang = response.get("language", None)
        duration = segments[-1]["end"] if segments else None
        # Process segments to extract or approximate word-level timestamps.
        words = self.segment_to_timestamped_words(segments)
        return TranscriptionResponse(
            text=text,
            segments=segments if include_segments else None,
            words=words,
            language=lang,
            duration=duration
        )

    def segment_to_timestamped_words(self, segments: List[dict]) -> List[TimeStampedWord]:
        """
        Converts a list of segments into word-level timestamps.
        If a segment includes a "words" key, it uses that data.
        Otherwise, it splits the segment text and assigns equal intervals.
        """
        output = []
        for segment in segments:
            # Skip segments that are likely silence.
            if segment.get("no_speech_prob", 0) > 0.9:
                continue
            if "words" in segment and segment["words"]:
                for word in segment["words"]:
                    output.append((word["start"], word["end"], word["word"]))
            else:
                # Fallback: split the segment text and assign approximate timestamps.
                words = segment["text"].split()
                if not words:
                    continue
                duration = segment["end"] - segment["start"]
                word_duration = duration / len(words)
                for i, w in enumerate(words):
                    start = segment["start"] + i * word_duration
                    end = start + word_duration
                    output.append((start, end, w))
        return output

    def segment_to_long_text(self, segments: List[Segment]) -> str:
        """
        Converts a list of segments into a full text.

        Args:
            segments (List[Segment]): Transcription segments with words and metadata.

        Returns:
            List[Tuple[float, float, str]]: List of (start, end, word) tuples.
        """

        output = []
        for segment in segments:
            if segment.no_speech_prob > 0.9:    # silence segment
                continue
            for word in segment.words:
                output.append(word.word)

        return self.sep.join(output)

    def segments_end_ts(self, segments: List[dict]) -> List[float]:
        """
        Returns a list of ending timestamps from the provided segments.
        """
        return [segment.get("end", 0.0) for segment in segments]

    def use_vad(self):
        """
        Enable voice activity detection (VAD) settings.
        For example, here we set a no-speech threshold in the extra keyword arguments.
        """
        self.transcribe_kargs["no_speech_threshold"] = 0.6

    def set_translate_task(self):
        self.transcribe_kargs["task"] = "translate"


# modified from https://github.com/lifeiteng/vall-e/blob/main/valle/modules/transformer.py
import copy
import numbers
from functools import partial
from typing import Any
from typing import Callable
from typing import List
from typing import Optional
from typing import Tuple
from typing import Union

import torch
from AR.modules.activation_onnx import MultiheadAttention
from AR.modules.scaling import BalancedDoubleSwish
from torch import nn
from torch import Tensor
from torch.nn import functional as F

_shape_t = Union[int, List[int], torch.Size]


class LayerNorm(nn.Module):
    __constants__ = ["normalized_shape", "eps", "elementwise_affine"]
    normalized_shape: Tuple[int, ...]
    eps: float
    elementwise_affine: bool

    def __init__(
        self,
        normalized_shape: _shape_t,
        eps: float = 1e-5,
        elementwise_affine: bool = True,
        device=None,
        dtype=None,
    ) -> None:
        factory_kwargs = {"device": device, "dtype": dtype}
        super(LayerNorm, self).__init__()
        if isinstance(normalized_shape, numbers.Integral):
            # mypy error: incompatible types in assignment
            normalized_shape = (normalized_shape,)  # type: ignore[assignment]
        self.normalized_shape = tuple(normalized_shape)  # type: ignore[arg-type]
        self.eps = eps
        self.elementwise_affine = elementwise_affine
        if self.elementwise_affine:
            self.weight = nn.Parameter(
                torch.empty(self.normalized_shape, **factory_kwargs)
            )
            self.bias = nn.Parameter(
                torch.empty(self.normalized_shape, **factory_kwargs)
            )
        else:
            self.register_parameter("weight", None)
            self.register_parameter("bias", None)

        self.reset_parameters()

    def reset_parameters(self) -> None:
        if self.elementwise_affine:
            nn.init.ones_(self.weight)
            nn.init.zeros_(self.bias)

    def forward(self, input: Tensor, embedding: Any = None) -> Tensor:
        if isinstance(input, tuple):
            input, embedding = input
            return (
                F.layer_norm(
                    input,
                    self.normalized_shape,
                    self.weight,
                    self.bias,
                    self.eps,
                ),
                embedding,
            )

        assert embedding is None
        return F.layer_norm(
            input, self.normalized_shape, self.weight, self.bias, self.eps
        )

    def extra_repr(self) -> str:
        return (
            "{normalized_shape}, eps={eps}, "
            "elementwise_affine={elementwise_affine}".format(**self.__dict__)
        )


class IdentityNorm(nn.Module):
    def __init__(
        self,
        d_model: int,
        eps: float = 1e-5,
        device=None,
        dtype=None,
    ) -> None:
        super(IdentityNorm, self).__init__()

    def forward(self, input: Tensor, embedding: Any = None) -> Tensor:
        if isinstance(input, tuple):
            return input

        assert embedding is None
        return input


class TransformerEncoder(nn.Module):
    r"""TransformerEncoder is a stack of N encoder layers. Users can build the
    BERT(https://arxiv.org/abs/1810.04805) model with corresponding parameters.

    Args:
        encoder_layer: an instance of the TransformerEncoderLayer() class (required).
        num_layers: the number of sub-encoder-layers in the encoder (required).
        norm: the layer normalization component (optional).
        enable_nested_tensor: if True, input will automatically convert to nested tensor
            (and convert back on output). This will improve the overall performance of
            TransformerEncoder when padding rate is high. Default: ``True`` (enabled).

    Examples::
        >>> encoder_layer = TransformerEncoderLayer(d_model=512, nhead=8)
        >>> transformer_encoder = TransformerEncoder(encoder_layer, num_layers=6)
        >>> src = torch.rand(10, 32, 512)
        >>> out = transformer_encoder(src)
    """
    __constants__ = ["norm"]

    def __init__(self, encoder_layer, num_layers, norm=None):
        super(TransformerEncoder, self).__init__()
        self.layers = _get_clones(encoder_layer, num_layers)
        self.num_layers = num_layers
        self.norm = norm

    def forward(
        self,
        src: Tensor,
        mask: Optional[Tensor] = None,
        src_key_padding_mask: Optional[Tensor] = None,
        return_layer_states: bool = False,
        cache=None,
    ) -> Tensor:
        output = src
        for mod in self.layers:
            output = mod(
                output,
                src_mask=mask,
                src_key_padding_mask=src_key_padding_mask,
                cache=cache,
            )

        if self.norm is not None:
            output = self.norm(output)

        return output


class TransformerEncoderLayer(nn.Module):
    __constants__ = ["batch_first", "norm_first"]
    def __init__(
        self,
        d_model: int,
        nhead: int,
        dim_feedforward: int = 2048,
        dropout: float = 0.1,
        activation: Union[str, Callable[[Tensor], Tensor]] = F.relu,
        batch_first: bool = False,
        norm_first: bool = False,
        device=None,
        dtype=None,
        linear1_self_attention_cls: nn.Module = nn.Linear,
        linear2_self_attention_cls: nn.Module = nn.Linear,
        linear1_feedforward_cls: nn.Module = nn.Linear,
        linear2_feedforward_cls: nn.Module = nn.Linear,
        layer_norm_cls: nn.Module = LayerNorm,
        layer_norm_eps: float = 1e-5,
        adaptive_layer_norm=False,
    ) -> None:
        factory_kwargs = {"device": device, "dtype": dtype}
        super(TransformerEncoderLayer, self).__init__()
        self.self_attn = MultiheadAttention(
            d_model,  # 512 16
            nhead,
            dropout=dropout,
            batch_first=batch_first,
            linear1_cls=linear1_self_attention_cls,
            linear2_cls=linear2_self_attention_cls,
            **factory_kwargs,
        )
        self.linear1 = linear1_feedforward_cls(
            d_model, dim_feedforward, **factory_kwargs
        )
        self.dropout = nn.Dropout(dropout)
        self.linear2 = linear2_feedforward_cls(
            dim_feedforward, d_model, **factory_kwargs
        )
        self.norm_first = norm_first
        self.dropout1 = nn.Dropout(dropout)
        self.dropout2 = nn.Dropout(dropout)
        if isinstance(activation, str):
            activation = _get_activation_fn(activation)
        elif isinstance(activation, partial):
            activation = activation(d_model)
        elif activation == BalancedDoubleSwish:
            activation = BalancedDoubleSwish(d_model)
        self.activation = activation

        norm1 = layer_norm_cls(d_model, eps=layer_norm_eps, **factory_kwargs)
        if layer_norm_cls == IdentityNorm:
            norm2 = BalancedBasicNorm(d_model, eps=layer_norm_eps, **factory_kwargs)
        else:
            norm2 = layer_norm_cls(d_model, eps=layer_norm_eps, **factory_kwargs)

        if adaptive_layer_norm:
            self.norm1 = AdaptiveLayerNorm(d_model, norm1)
            self.norm2 = AdaptiveLayerNorm(d_model, norm2)
        else:
            self.norm1 = norm1
            self.norm2 = norm2

    def __setstate__(self, state):
        super(TransformerEncoderLayer, self).__setstate__(state)
        if not hasattr(self, "activation"):
            self.activation = F.relu

    def forward(
        self,
        src: Tensor,
        src_mask: Optional[Tensor] = None,
        src_key_padding_mask: Optional[Tensor] = None,
        cache=None,
    ) -> Tensor:
        x = src
        stage_embedding = None
        x = self.norm1(
            x + self._sa_block(x, src_mask, src_key_padding_mask, cache=cache),
            stage_embedding,
        )
        x = self.norm2(x + self._ff_block(x), stage_embedding)

        return x

    def _sa_block(
        self,
        x: Tensor,
        attn_mask: Optional[Tensor],
        key_padding_mask: Optional[Tensor],
        cache=None,
    ) -> Tensor:
        x = self.self_attn(
            x,
            x,
            x,
            attn_mask=attn_mask,
            key_padding_mask=key_padding_mask,
            need_weights=False,
            cache=cache,
        )
        return self.dropout1(x)

    def _ff_block(self, x: Tensor) -> Tensor:
        x = self.linear2(self.dropout(self.activation(self.linear1(x))))
        return self.dropout2(x)


class AdaptiveLayerNorm(nn.Module):
    r"""Adaptive Layer Normalization"""

    def __init__(self, d_model, norm) -> None:
        super(AdaptiveLayerNorm, self).__init__()
        self.project_layer = nn.Linear(d_model, 2 * d_model)
        self.norm = norm
        self.d_model = d_model
        self.eps = self.norm.eps

    def forward(self, input: Tensor, embedding: Tensor = None) -> Tensor:
        if isinstance(input, tuple):
            input, embedding = input
            weight, bias = torch.split(
                self.project_layer(embedding),
                split_size_or_sections=self.d_model,
                dim=-1,
            )
            return (weight * self.norm(input) + bias, embedding)

        weight, bias = torch.split(
            self.project_layer(embedding),
            split_size_or_sections=self.d_model,
            dim=-1,
        )
        return weight * self.norm(input) + bias


def _get_clones(module, N):
    return nn.ModuleList([copy.deepcopy(module) for i in range(N)])



import os,sys
parent_directory = os.path.dirname(os.path.abspath(__file__))
import logging,pdb
logger = logging.getLogger(__name__)

import librosa
import numpy as np
import soundfile as sf
import torch
from lib.lib_v5 import nets_61968KB as Nets
from lib.lib_v5 import spec_utils
from lib.lib_v5.model_param_init import ModelParameters
from lib.lib_v5.nets_new import CascadedNet
from lib.utils import inference


class AudioPre:
    def __init__(self, agg, model_path, device, is_half, tta=False):
        self.model_path = model_path
        self.device = device
        self.data = {
            # Processing Options
            "postprocess": False,
            "tta": tta,
            # Constants
            "window_size": 512,
            "agg": agg,
            "high_end_process": "mirroring",
        }
        mp = ModelParameters("%s/lib/lib_v5/modelparams/4band_v2.json"%parent_directory)
        model = Nets.CascadedASPPNet(mp.param["bins"] * 2)
        cpk = torch.load(model_path, map_location="cpu")
        model.load_state_dict(cpk)
        model.eval()
        if is_half:
            model = model.half().to(device)
        else:
            model = model.to(device)

        self.mp = mp
        self.model = model

    def _path_audio_(
        self, music_file, ins_root=None, vocal_root=None, format="flac", is_hp3=False
    ):
        if ins_root is None and vocal_root is None:
            return "No save root."
        name = os.path.basename(music_file)
        if ins_root is not None:
            os.makedirs(ins_root, exist_ok=True)
        if vocal_root is not None:
            os.makedirs(vocal_root, exist_ok=True)
        X_wave, y_wave, X_spec_s, y_spec_s = {}, {}, {}, {}
        bands_n = len(self.mp.param["band"])
        # print(bands_n)
        for d in range(bands_n, 0, -1):
            bp = self.mp.param["band"][d]
            if d == bands_n:  # high-end band
                (
                    X_wave[d],
                    _,
                ) = librosa.core.load(  # 理论上librosa读取可能对某些音频有bug，应该上ffmpeg读取，但是太麻烦了弃坑
                    music_file,
                    sr       = bp["sr"],
                    mono     = False,
                    dtype    = np.float32,
                    res_type = bp["res_type"],
                )
                if X_wave[d].ndim == 1:
                    X_wave[d] = np.asfortranarray([X_wave[d], X_wave[d]])
            else:  # lower bands
                X_wave[d] = librosa.core.resample(
                    X_wave[d + 1],
                    orig_sr   = self.mp.param["band"][d + 1]["sr"],
                    target_sr = bp["sr"],
                    res_type  = bp["res_type"],
                )
            # Stft of wave source
            X_spec_s[d] = spec_utils.wave_to_spectrogram_mt(
                X_wave[d],
                bp["hl"],
                bp["n_fft"],
                self.mp.param["mid_side"],
                self.mp.param["mid_side_b2"],
                self.mp.param["reverse"],
            )
            # pdb.set_trace()
            if d == bands_n and self.data["high_end_process"] != "none":
                input_high_end_h = (bp["n_fft"] // 2 - bp["crop_stop"]) + (
                    self.mp.param["pre_filter_stop"] - self.mp.param["pre_filter_start"]
                )
                input_high_end = X_spec_s[d][
                    :, bp["n_fft"] // 2 - input_high_end_h : bp["n_fft"] // 2, :
                ]

        X_spec_m = spec_utils.combine_spectrograms(X_spec_s, self.mp)
        aggresive_set = float(self.data["agg"] / 100)
        aggressiveness = {
            "value": aggresive_set,
            "split_bin": self.mp.param["band"][1]["crop_stop"],
        }
        with torch.no_grad():
            pred, X_mag, X_phase = inference(
                X_spec_m, self.device, self.model, aggressiveness, self.data
            )
        # Postprocess
        if self.data["postprocess"]:
            pred_inv = np.clip(X_mag - pred, 0, np.inf)
            pred = spec_utils.mask_silence(pred, pred_inv)
        y_spec_m = pred * X_phase
        v_spec_m = X_spec_m - y_spec_m

        if is_hp3 == True:
            ins_root,vocal_root = vocal_root,ins_root

        if ins_root is not None:
            if self.data["high_end_process"].startswith("mirroring"):
                input_high_end_ = spec_utils.mirroring(
                    self.data["high_end_process"], y_spec_m, input_high_end, self.mp
                )
                wav_instrument = spec_utils.cmb_spectrogram_to_wave(
                    y_spec_m, self.mp, input_high_end_h, input_high_end_
                )
            else:
                wav_instrument = spec_utils.cmb_spectrogram_to_wave(y_spec_m, self.mp)
            logger.info("%s instruments done" % name)
            if is_hp3 == True:
                head = "vocal_"
            else:
                head = "instrument_"
            if format in ["wav", "flac"]:
                sf.write(
                    os.path.join(
                        ins_root,
                        head + "{}_{}.{}".format(name, self.data["agg"], format),
                    ),
                    (np.array(wav_instrument) * 32768).astype("int16"),
                    self.mp.param["sr"],
                )  #
            else:
                path = os.path.join(
                    ins_root, head + "{}_{}.wav".format(name, self.data["agg"])
                )
                sf.write(
                    path,
                    (np.array(wav_instrument) * 32768).astype("int16"),
                    self.mp.param["sr"],
                )
                if os.path.exists(path):
                    opt_format_path = path[:-4] + ".%s" % format
                    os.system("ffmpeg -i %s -vn %s -q:a 2 -y" % (path, opt_format_path))
                    if os.path.exists(opt_format_path):
                        try:
                            os.remove(path)
                        except:
                            pass
        if vocal_root is not None:
            if is_hp3 == True:
                head = "instrument_"
            else:
                head = "vocal_"
            if self.data["high_end_process"].startswith("mirroring"):
                input_high_end_ = spec_utils.mirroring(
                    self.data["high_end_process"], v_spec_m, input_high_end, self.mp
                )
                wav_vocals = spec_utils.cmb_spectrogram_to_wave(
                    v_spec_m, self.mp, input_high_end_h, input_high_end_
                )
            else:
                wav_vocals = spec_utils.cmb_spectrogram_to_wave(v_spec_m, self.mp)
            logger.info("%s vocals done" % name)
            if format in ["wav", "flac"]:
                sf.write(
                    os.path.join(
                        vocal_root,
                        head + "{}_{}.{}".format(name, self.data["agg"], format),
                    ),
                    (np.array(wav_vocals) * 32768).astype("int16"),
                    self.mp.param["sr"],
                )
            else:
                path = os.path.join(
                    vocal_root, head + "{}_{}.wav".format(name, self.data["agg"])
                )
                sf.write(
                    path,
                    (np.array(wav_vocals) * 32768).astype("int16"),
                    self.mp.param["sr"],
                )
                if os.path.exists(path):
                    opt_format_path = path[:-4] + ".%s" % format
                    os.system("ffmpeg -i %s -vn %s -q:a 2 -y" % (path, opt_format_path))
                    if os.path.exists(opt_format_path):
                        try:
                            os.remove(path)
                        except:
                            pass


class AudioPreDeEcho:
    def __init__(self, agg, model_path, device, is_half, tta=False):
        self.model_path = model_path
        self.device = device
        self.data = {
            # Processing Options
            "postprocess": False,
            "tta": tta,
            # Constants
            "window_size": 512,
            "agg": agg,
            "high_end_process": "mirroring",
        }
        mp = ModelParameters("%s/lib/lib_v5/modelparams/4band_v3.json"%parent_directory)
        nout = 64 if "DeReverb" in model_path else 48
        model = CascadedNet(mp.param["bins"] * 2, nout)
        cpk = torch.load(model_path, map_location="cpu")
        model.load_state_dict(cpk)
        model.eval()
        if is_half:
            model = model.half().to(device)
        else:
            model = model.to(device)

        self.mp = mp
        self.model = model

    def _path_audio_(
        self, music_file, vocal_root=None, ins_root=None, format="flac", is_hp3=False
    ):  # 3个VR模型vocal和ins是反的
        if ins_root is None and vocal_root is None:
            return "No save root."
        name = os.path.basename(music_file)
        if ins_root is not None:
            os.makedirs(ins_root, exist_ok=True)
        if vocal_root is not None:
            os.makedirs(vocal_root, exist_ok=True)
        X_wave, y_wave, X_spec_s, y_spec_s = {}, {}, {}, {}
        bands_n = len(self.mp.param["band"])
        # print(bands_n)
        for d in range(bands_n, 0, -1):
            bp = self.mp.param["band"][d]
            if d == bands_n:  # high-end band
                (
                    X_wave[d],
                    _,
                ) = librosa.core.load(  # 理论上librosa读取可能对某些音频有bug，应该上ffmpeg读取，但是太麻烦了弃坑
                    music_file,
                    sr       = bp["sr"],
                    mono     = False,
                    dtype    = np.float32,
                    res_type = bp["res_type"],
                )
                if X_wave[d].ndim == 1:
                    X_wave[d] = np.asfortranarray([X_wave[d], X_wave[d]])
            else:  # lower bands
                X_wave[d] = librosa.core.resample(
                    X_wave[d + 1],
                    orig_sr   = self.mp.param["band"][d + 1]["sr"],
                    target_sr = bp["sr"],
                    res_type  = bp["res_type"],
                )
            # Stft of wave source
            X_spec_s[d] = spec_utils.wave_to_spectrogram_mt(
                X_wave[d],
                bp["hl"],
                bp["n_fft"],
                self.mp.param["mid_side"],
                self.mp.param["mid_side_b2"],
                self.mp.param["reverse"],
            )
            # pdb.set_trace()
            if d == bands_n and self.data["high_end_process"] != "none":
                input_high_end_h = (bp["n_fft"] // 2 - bp["crop_stop"]) + (
                    self.mp.param["pre_filter_stop"] - self.mp.param["pre_filter_start"]
                )
                input_high_end = X_spec_s[d][
                    :, bp["n_fft"] // 2 - input_high_end_h : bp["n_fft"] // 2, :
                ]

        X_spec_m = spec_utils.combine_spectrograms(X_spec_s, self.mp)
        aggresive_set = float(self.data["agg"] / 100)
        aggressiveness = {
            "value": aggresive_set,
            "split_bin": self.mp.param["band"][1]["crop_stop"],
        }
        with torch.no_grad():
            pred, X_mag, X_phase = inference(
                X_spec_m, self.device, self.model, aggressiveness, self.data
            )
        # Postprocess
        if self.data["postprocess"]:
            pred_inv = np.clip(X_mag - pred, 0, np.inf)
            pred = spec_utils.mask_silence(pred, pred_inv)
        y_spec_m = pred * X_phase
        v_spec_m = X_spec_m - y_spec_m

        if ins_root is not None:
            if self.data["high_end_process"].startswith("mirroring"):
                input_high_end_ = spec_utils.mirroring(
                    self.data["high_end_process"], y_spec_m, input_high_end, self.mp
                )
                wav_instrument = spec_utils.cmb_spectrogram_to_wave(
                    y_spec_m, self.mp, input_high_end_h, input_high_end_
                )
            else:
                wav_instrument = spec_utils.cmb_spectrogram_to_wave(y_spec_m, self.mp)
            logger.info("%s instruments done" % name)
            if format in ["wav", "flac"]:
                sf.write(
                    os.path.join(
                        ins_root,
                        "vocal_{}_{}.{}".format(name, self.data["agg"], format),
                    ),
                    (np.array(wav_instrument) * 32768).astype("int16"),
                    self.mp.param["sr"],
                )  #
            else:
                path = os.path.join(
                    ins_root, "vocal_{}_{}.wav".format(name, self.data["agg"])
                )
                sf.write(
                    path,
                    (np.array(wav_instrument) * 32768).astype("int16"),
                    self.mp.param["sr"],
                )
                if os.path.exists(path):
                    opt_format_path = path[:-4] + ".%s" % format
                    os.system("ffmpeg -i %s -vn %s -q:a 2 -y" % (path, opt_format_path))
                    if os.path.exists(opt_format_path):
                        try:
                            os.remove(path)
                        except:
                            pass
        if vocal_root is not None:
            if self.data["high_end_process"].startswith("mirroring"):
                input_high_end_ = spec_utils.mirroring(
                    self.data["high_end_process"], v_spec_m, input_high_end, self.mp
                )
                wav_vocals = spec_utils.cmb_spectrogram_to_wave(
                    v_spec_m, self.mp, input_high_end_h, input_high_end_
                )
            else:
                wav_vocals = spec_utils.cmb_spectrogram_to_wave(v_spec_m, self.mp)
            logger.info("%s vocals done" % name)
            if format in ["wav", "flac"]:
                sf.write(
                    os.path.join(
                        vocal_root,
                        "instrument_{}_{}.{}".format(name, self.data["agg"], format),
                    ),
                    (np.array(wav_vocals) * 32768).astype("int16"),
                    self.mp.param["sr"],
                )
            else:
                path = os.path.join(
                    vocal_root, "instrument_{}_{}.wav".format(name, self.data["agg"])
                )
                sf.write(
                    path,
                    (np.array(wav_vocals) * 32768).astype("int16"),
                    self.mp.param["sr"],
                )
                if os.path.exists(path):
                    opt_format_path = path[:-4] + ".%s" % format
                    os.system("ffmpeg -i %s -vn %s -q:a 2 -y" % (path, opt_format_path))
                    if os.path.exists(opt_format_path):
                        try:
                            os.remove(path)
                        except:
                            pass



from typing import List, Tuple
from gallama.logger.logger import logger
from ...data_classes import TimeStampedWord


class HypothesisBuffer:

    def __init__(self):
        """
        This class manages hypotheses (potential recognized words) from the ASR system and determines which words can be committed as finalized transcriptions.
        """
        self.commited_in_buffer = []  # Stores finalized (committed) transcriptions.
        self.buffer = []  # Stores new hypotheses inserted but not yet processed.
        self.new: List[Tuple[float, float, str]] = []  # Stores new hypotheses inserted but not yet processed.

        self.last_commited_time = 0  # Timestamp of the last committed word.
        self.last_commited_word = None  # Last committed word.
        self.is_final = False  # New flag for final processing


        # Parameters for detecting repetition patterns
        self.min_word_gap = 0.1  # Minimum expected gap between distinct words
        self.max_merge_window = 2.0  # Maximum window to consider for merging

    def detect_repetition_pattern(self, words):
        """
        Analyze sequence of words to detect abnormal repetition patterns
        Returns list of indices that are likely part of a repetition pattern
        """
        if len(words) < 2:
            return []

        repetition_indices = []
        for i in range(1, len(words)):
            prev_time = words[i - 1][1]  # End time of previous word
            curr_time = words[i][0]  # Start time of current word
            time_gap = curr_time - prev_time

            # Check for suspiciously small gaps between words
            if time_gap < self.min_word_gap:
                # Look for repetition of single characters or very short segments
                if len(words[i][2]) <= 2 and words[i][2] == words[i - 1][2]:
                    repetition_indices.extend([i - 1, i])

        return list(set(repetition_indices))

    def merge_adjacent_segments(self, words):
        """
        Merge adjacent word segments that might be part of the same word
        """
        if len(words) < 2:
            return words

        merged = []
        i = 0
        while i < len(words):
            if i == len(words) - 1:
                merged.append(words[i])
                break

            curr_word = words[i]
            next_word = words[i + 1]

            # Check if these segments should be merged
            time_gap = next_word[0] - curr_word[1]
            if time_gap < self.min_word_gap and len(curr_word[2]) <= 2 and len(next_word[2]) <= 2:
                # Merge the segments
                merged_text = curr_word[2] + next_word[2]
                merged_segment = (curr_word[0], next_word[1], merged_text)
                merged.append(merged_segment)
                i += 2
            else:
                merged.append(curr_word)
                i += 1

        return merged

    def insert(self, new: TimeStampedWord, offset: float):  # TODO
        """
        Adjusts the timestamps of the new hypotheses by offset.
        Filters out any hypotheses that are older than last_commited_time - 0.1.
        Removes overlapping words between commited_in_buffer and new (using n-gram matching).

        How the algorithm works:
        compare self.commited_in_buffer and new. It inserts only the words in new that extend the commited_in_buffer, it means they are roughly behind last_commited_time and new in content
        the new tail is added to self.new
        """

        # Apply offset to timestamps
        new = [(a + offset, b + offset, t) for a, b, t in new]

        # Filter based on timestamp
        filtered_new = [(a, b, t) for a, b, t in new if a > self.last_commited_time - 0.5]

        if filtered_new:
            # Detect repetition patterns
            rep_indices = self.detect_repetition_pattern(filtered_new)

            # Remove detected repetitions
            filtered_new = [word for i, word in enumerate(filtered_new)
                            if i not in rep_indices]

            # Merge adjacent segments that might be part of the same word
            filtered_new = self.merge_adjacent_segments(filtered_new)

            self.new = filtered_new

            # Perform n-gram matching with existing buffer
            if self.new and abs(self.new[0][0] - self.last_commited_time) < 1:
                if self.commited_in_buffer:
                    cn = len(self.commited_in_buffer)
                    nn = len(self.new)
                    for i in range(1, min(min(cn, nn), 5) + 1):
                        c = " ".join([self.commited_in_buffer[-j][2]
                                      for j in range(1, i + 1)][::-1])
                        tail = " ".join(self.new[j - 1][2]
                                        for j in range(1, i + 1))
                        if c == tail:
                            for j in range(i):
                                self.new.pop(0)
                            break

    def flush(self, force_commit=False):
        """
        Commits words to the final transcription.
        Args:
            force_commit: If True, commits remaining words without waiting for confirmation
        """
        commit = []

        if force_commit:
            # Commit all remaining words in buffer and new
            remaining_words = []
            if self.buffer:
                remaining_words.extend(self.buffer)
            if self.new:
                remaining_words.extend(self.new)

            # Sort by start time
            remaining_words.sort(key=lambda x: x[0])

            # Remove duplicates while preserving order
            seen_words = {}  # Using dict to preserve order
            for na, nb, nt in remaining_words:
                # If we see the same word again, keep the one with the later timestamp
                if nt in seen_words:
                    prev_na, prev_nb, _ = seen_words[nt]
                    if nb > prev_nb:
                        seen_words[nt] = (na, nb, nt)
                else:
                    seen_words[nt] = (na, nb, nt)

            # Convert back to list
            commit = list(seen_words.values())

            # Update last committed info
            if commit:
                self.last_commited_word = commit[-1][2]
                self.last_commited_time = commit[-1][1]

            self.buffer = []
            self.new = []

        else:
            # Normal matching-based commitment
            while self.new and self.buffer:
                na, nb, nt = self.new[0]
                if nt == self.buffer[0][2]:
                    commit.append((na, nb, nt))
                    self.last_commited_word = nt
                    self.last_commited_time = nb
                    self.buffer.pop(0)
                    self.new.pop(0)
                else:
                    break

            self.buffer = self.new
            self.new = []

        self.commited_in_buffer.extend(commit)
        return commit

    def set_final(self):
        """Mark the buffer as final, triggering complete word commitment"""
        self.is_final = True
        return self.flush(force_commit=True)

    def pop_committed(self, time):
        while self.commited_in_buffer and self.commited_in_buffer[0][1] <= time:
            self.commited_in_buffer.pop(0)

    def complete(self):
        """Returns all remaining words in the buffer"""
        return self.buffer



# -*- coding: utf-8 -*-

import os

inp_text = os.environ.get("inp_text")
inp_wav_dir = os.environ.get("inp_wav_dir")
exp_name = os.environ.get("exp_name")
i_part = os.environ.get("i_part")
all_parts = os.environ.get("all_parts")
if "_CUDA_VISIBLE_DEVICES" in os.environ:
     os.environ["CUDA_VISIBLE_DEVICES"] = os.environ["_CUDA_VISIBLE_DEVICES"]
opt_dir = os.environ.get("opt_dir")
bert_pretrained_dir = os.environ.get("bert_pretrained_dir")
import torch
is_half = eval(os.environ.get("is_half", "True")) and torch.cuda.is_available()
version = os.environ.get('version', None)
import sys, numpy as np, traceback, pdb
import os.path
from glob import glob
from tqdm import tqdm
from text.cleaner import clean_text
from transformers import AutoModelForMaskedLM, AutoTokenizer
import numpy as np
from tools.my_utils import clean_path

# inp_text=sys.argv[1]
# inp_wav_dir=sys.argv[2]
# exp_name=sys.argv[3]
# i_part=sys.argv[4]
# all_parts=sys.argv[5]
# os.environ["CUDA_VISIBLE_DEVICES"]=sys.argv[6]#i_gpu
# opt_dir="/data/docker/liujing04/gpt-vits/fine_tune_dataset/%s"%exp_name
# bert_pretrained_dir="/data/docker/liujing04/bert-vits2/Bert-VITS2-master20231106/bert/chinese-roberta-wwm-ext-large"

from time import time as ttime
import shutil


def my_save(fea,path):#####fix issue: torch.save doesn't support chinese path
    dir=os.path.dirname(path)
    name=os.path.basename(path)
    # tmp_path="%s/%s%s.pth"%(dir,ttime(),i_part)
    tmp_path="%s%s.pth"%(ttime(),i_part)
    torch.save(fea,tmp_path)
    shutil.move(tmp_path,"%s/%s"%(dir,name))


txt_path = "%s/2-name2text-%s.txt" % (opt_dir, i_part)
if os.path.exists(txt_path) == False:
    bert_dir = "%s/3-bert" % (opt_dir)
    os.makedirs(opt_dir, exist_ok=True)
    os.makedirs(bert_dir, exist_ok=True)
    if torch.cuda.is_available():
        device = "cuda:0"
    # elif torch.backends.mps.is_available():
    #     device = "mps"
    else:
        device = "cpu"
    if os.path.exists(bert_pretrained_dir):...
    else:raise FileNotFoundError(bert_pretrained_dir)
    tokenizer = AutoTokenizer.from_pretrained(bert_pretrained_dir)
    bert_model = AutoModelForMaskedLM.from_pretrained(bert_pretrained_dir)
    if is_half == True:
        bert_model = bert_model.half().to(device)
    else:
        bert_model = bert_model.to(device)

    def get_bert_feature(text, word2ph):
        with torch.no_grad():
            inputs = tokenizer(text, return_tensors="pt")
            for i in inputs:
                inputs[i] = inputs[i].to(device)
            res = bert_model(**inputs, output_hidden_states=True)
            res = torch.cat(res["hidden_states"][-3:-2], -1)[0].cpu()[1:-1]

        assert len(word2ph) == len(text)
        phone_level_feature = []
        for i in range(len(word2ph)):
            repeat_feature = res[i].repeat(word2ph[i], 1)
            phone_level_feature.append(repeat_feature)

        phone_level_feature = torch.cat(phone_level_feature, dim=0)

        return phone_level_feature.T

    def process(data, res):
        for name, text, lan in data:
            try:
                name=clean_path(name)
                name = os.path.basename(name)
                print(name)
                phones, word2ph, norm_text = clean_text(
                    text.replace("%", "-").replace("￥", ","), lan, version
                )
                path_bert = "%s/%s.pt" % (bert_dir, name)
                if os.path.exists(path_bert) == False and lan == "zh":
                    bert_feature = get_bert_feature(norm_text, word2ph)
                    assert bert_feature.shape[-1] == len(phones)
                    # torch.save(bert_feature, path_bert)
                    my_save(bert_feature, path_bert)
                phones = " ".join(phones)
                # res.append([name,phones])
                res.append([name, phones, word2ph, norm_text])
            except:
                print(name, text, traceback.format_exc())

    todo = []
    res = []
    with open(inp_text, "r", encoding="utf8") as f:
        lines = f.read().strip("\n").split("\n")

    language_v1_to_language_v2 = {
        "ZH": "zh",
        "zh": "zh",
        "JP": "ja",
        "jp": "ja",
        "JA": "ja",
        "ja": "ja",
        "EN": "en",
        "en": "en",
        "En": "en",
        "KO": "ko",
        "Ko": "ko",
        "ko": "ko",
        "yue": "yue",
        "YUE": "yue",
        "Yue": "yue",
    }
    for line in lines[int(i_part) :: int(all_parts)]:
        try:
            wav_name, spk_name, language, text = line.split("|")
            # todo.append([name,text,"zh"])
            if language in language_v1_to_language_v2.keys():
                todo.append(
                    [wav_name, text, language_v1_to_language_v2.get(language, language)]
                )
            else:
                print(f"\033[33m[Waring] The {language = } of {wav_name} is not supported for training.\033[0m")
        except:
            print(line, traceback.format_exc())

    process(todo, res)
    opt = []
    for name, phones, word2ph, norm_text in res:
        opt.append("%s\t%s\t%s\t%s" % (name, phones, word2ph, norm_text))
    with open(txt_path, "w", encoding="utf8") as f:
        f.write("\n".join(opt) + "\n")



from abc import ABC, abstractmethod
from typing import Dict, List, Union, Optional, Literal, Callable, Tuple
import asyncio
import re       # for text processing of the thinking
from fastapi import HTTPException, Request
import textwrap

# logger
from gallama.logger import logger

# format enforcement
from lmformatenforcer.tokenenforcer import TokenEnforcerTokenizerData
from formatron.formatter import FormatterBuilder
from formatron.schemas.pydantic import ClassSchema
from gallama.backend.llm.format_enforcer import FormatEnforcer

# thinking
from gallama.backend.llm.thinking_template import THINKING_TEMPLATE, Thinking

# function calling
from gallama.backend.llm.tools import Tools, create_function_models_v2, create_function_models_formatron

from gallama.utils.utils import get_token_length
from gallama.api_response.chat_response import get_response_from_queue   # helper function to collect result from queue
from gallama.data_classes import (
    ModelSpec,
    ChatMLQuery,
    GenEnd,
    GenText,
    GenQueue,
    GenStart,
    GenerationStats,
    QueueContext,
    GenQueueDynamic,
    VideoFrame,
    MultiModalTextContent,
    MultiModalImageContent
)
from ....config.config_manager import ConfigManager
# handle prompting
from gallama.backend.llm.prompt_engine import PromptEngine
from dataclasses import dataclass

# video handling
from ....data_classes import VideoFrameCollection

config_manager = ConfigManager()

@dataclass
class ToolCallV2:
    gen_dynamic_queue: List[GenQueueDynamic]
    stop_event: asyncio.Event
    generate_kwargs: dict

class ModelInterface(ABC):
    @abstractmethod
    def __init__(self, model_spec:ModelSpec):
        # initialization share the same code to keep the frontend consistent
        # if a backend does not support the value, please set it in the __init__ or load_model()

        # load prompt engine
        self.prompt_eng = PromptEngine(prompt_format=model_spec.prompt_template)

        # model_spec capture cli argument
        # model_config is from yml file
        self.model_id = model_spec.model_id
        # self.model_name = model_spec.model_name or model_config["model_name"]
        self.model_name = model_spec.model_name
        # self.max_seq_len = model_spec.max_seq_len or model_config.get("max_seq_len", None)
        self.max_seq_len = model_spec.max_seq_len
        if self.max_seq_len is not None:
            self.max_seq_len = (self.max_seq_len//256) * 256     # for paged attention

        # self.gpus = model_spec.gpus or model_config.get("gpus") or "auto"
        self.gpus = model_spec.gpus or "auto"
        # self.cache_size = model_spec.cache_size or model_config.get("cache_size") or self.max_seq_len   # default to max_seq_len if not set
        self.cache_size = model_spec.cache_size or self.max_seq_len   # default to max_seq_len if not set
        if self.cache_size is not None:
            self.cache_size = (self.cache_size//256) * 256     # for paged attention
            if self.max_seq_len is not None:
                # cache size must be greater or equal to max_seq_len
                self.cache_size = max(self.cache_size, self.max_seq_len)

        # self.cache_quant = model_spec.cache_quant or model_config.get("cache_quant") or "Q4"
        self.cache_quant = model_spec.cache_quant or "Q6"       # default to cache quant 6
        # self.backend = model_spec.backend or model_config["backend"] or "exllama"
        self.backend = model_spec.backend   # default should be set as exllama if not defined
        # self.tensor_parallel = model_spec.tensor_parallel or model_config.get("tensor_parallel", False)
        self.tensor_parallel = model_spec.tensor_parallel or False      # tensor parallel is False unless explicitly

        # transformers specific arguments
        # self.backend_extra_args = model_spec.get("backend_extra_args") or {}
        self.backend_extra_args = model_spec.backend_extra_args or {}


        # handle draft model
        draft_model_config = {}
        if model_spec.draft_model_id:
            draft_model_config = config_manager.get_model_config(model_spec.draft_model_name)
            if not draft_model_config:
                raise HTTPException(f"Model config for '{model_spec.draft_model_name}' not exist")


        # draft model is via cli only
        self.draft_model_id = draft_model_config.get("model_id")
        self.draft_model_name = model_spec.draft_model_name or None
        self.draft_gpus = model_spec.draft_gpus or draft_model_config.get("draft_gpus") or "auto"
        self.draft_cache_size = self.cache_size   # set to the same as main model
        self.draft_cache_quant = model_spec.draft_cache_quant or draft_model_config.get("cache_quant") or "Q4"
        # assert (self.draft_model_id is None) == (self.draft_model_name is None)

        # get the eos_token_str by merging the default config with anything set by user
        self.eos_token_str = list(set(model_spec.eos_token_list + self.prompt_eng.eos_token_list))
        self.eos_token_str_set = set(self.eos_token_str)    # set for some more efficient operation

        # load_model method in each subclass should set the following parameters:
        self.model = None
        self.tokenizer = None
        self.cache = None
        self.draft_model = None
        self.draft_cache = None
        self.processor = None       # processor is for processing of vision input
        self.eos_token_ids = None

        # placeholder
        # pipeline is a wrapper for model so that generation can be more standardized
        # this is because each backend expose different interface
        self.pipeline = None

        # format enforcer
        self.formatter = FormatEnforcer()

        # standard tool template
        self.TOOL_THINKING = THINKING_TEMPLATE["tool_necessity_evaluation"]
        self.TOOL_FORCE_THINKING = THINKING_TEMPLATE["tool_forced_evaluation"]

        # standard modalities
        # this set should contain either "text", "image", "video"
        self.modalities = {"text"}     # TODO current backend just assume that image is supported

    ## *************** the following method must be implemented by each backend ********
    @property
    def support_concurrency(self) -> bool:
        """
        whether this backend/ model support concurrent request
        """
        return False

    @property
    def support_tool(self) -> bool:
        """
        whether this backend/ model support format enforcement for tool
        """
        return True

    @abstractmethod
    def load_model(self):
        """Load the model, tokenizer, cache, and optional processor."""
        pass

    @abstractmethod
    def generate_eos_tokens_id(self) -> List[int]:
        """Generate the end-of-sequence token IDs."""
        pass

    @abstractmethod
    async def generate(
        self,
        prompt: str,
        gen_queue: Union[GenQueue, QueueContext, List[QueueContext]],
        request: Optional[Request] = None,
        gen_type: Union[str, GenStart] = "text", # the generated result will be store to this queue
        temperature: float = 0.01,
        top_p: float = 0.8,
        formatter: FormatterBuilder | TokenEnforcerTokenizerData = None,
        stop_words: Union[List[str], str] = None,
        prefix_strings: Optional[Union[str, List[str]]] = None,
        banned_strings: list[str] | None = None,
        max_tokens: int = None,
        quiet: bool = False,
        messages: List = None,  # query.message is used for multimodal
        video: List[VideoFrame] = None,
        stop_event: asyncio.Event = None,
        **kwargs,
    ) -> (str, GenerationStats):
        pass

    @property
    def support_video(self) -> bool:
        """
        whether this backend/ model support concurrent request
        """
        return "video" in self.modalities

    @property
    def support_format_enforcer(self) -> bool:
        return True

    @property
    def video_token_by_backend(self) -> str:
        return ""

    ## ******************************************************************************
    def validate_video_support(self, query: ChatMLQuery):
        """
        Validate if the current model have video support.
        Convert video frame to base64 images if the current model does not support video.
        The converted images will be attached to the last message's content.
        After conversion, the query.video is cleared.
        """
        # If no video frames exist, nothing to do.
        # If no messages, nothing to do
        # If already support video, nothing to do
        if not query.video or not query.messages or self.support_video:
            # add in video token if there is video
            if query.video and self.video_token_by_backend:
                last_message = query.messages[-1]

                if isinstance(last_message.content, str):
                    last_message.content += self.video_token_by_backend
                elif isinstance(last_message.content, list):
                    last_message.content.append(
                        MultiModalTextContent(type="text", text=self.video_token_by_backend)
                    )

            return query

        # raise error if image is not support
        if query.video and "image" not in self.modalities:
            raise Exception("Error while generating chat response: " + str(e))

        # Convert all video frames to base64-encoded PNG Data URIs.
        # We assume query.video is a list of VideoFrame objects.
        logger.debug("Converting video frame to base64 images")
        base64_images = VideoFrameCollection.convert_frames_to_base64(query.video)

        # If there are messages, update the last message.
        last_message = query.messages[-1]

        # If the content is a string, convert it into a list.
        if isinstance(last_message.content, str):
            # Preserve existing text if any.
            if last_message.content:
                last_message.content = [MultiModalTextContent(type="text", text=last_message.content)]
            else:
                last_message.content = []

        # Append each base64 image as a new image content.
        for img in base64_images:
            last_message.content.append(
                MultiModalImageContent(
                    type="image_url",
                    image_url=MultiModalImageContent.ImageDetail(
                        url=img,
                        detail="high"
                    )
                )
            )

        # Remove video frames from the query.
        query.video = None

        return query

    async def chat(
        self,
        query: ChatMLQuery,
        prompt_eng: PromptEngine,
        gen_queue: GenQueue,
        request: Request,
        stop_event: asyncio.Event = asyncio.Event()
    ):
        """
        This function will route the request to chat with tool or without tool accordingly
        """

        query = self.validate_video_support(query)

        # set the suitable method for tool calling
        if (query.tools or query.tool_choice != "none") and self.support_tool:
            chat_method = self.chat_with_tool_v2
        else:
            # enforce tool_choice to none if it is backend issue
            if not self.support_tool:
                logger.info("Tool not supported for this backend")
                query.tool_choice = "none"

            chat_method = self.chat_no_tool

        try:
            await chat_method(
                query=query,
                prompt_eng=prompt_eng,
                gen_queue=gen_queue,
                request=request,
                stop_event=stop_event,
            )
        except Exception as e:
            if stop_event.is_set():
                logger.info("Stop event is set, aborting chat")
            else:
                logger.error("Error while generating response: " + str(e))
                raise Exception("Error while generating chat response: " + str(e))
        return True

    async def chat_raw(
        self,
        prompt: str,
        gen_queue: asyncio.Queue,
        request: Request,
        # stream: bool = False,
        max_tokens: int = None,
        quiet=False,    # to disable any logging, mostly used for initialization cache prefill
        stop_event: asyncio.Event = asyncio.Event()
    ):
        """
        This function handle chat with input as a string prompt.
        This is mostly used for internal generation of the engine where the input is only a string
        """
        return await self.generate(
            prompt,
            max_tokens=max_tokens,
            gen_queue=gen_queue,
            quiet=quiet,
            request=request,
            stop_event=stop_event,
        )

    def validate_token_length(self, token_length: int):
        """
        validate that token_length is within max sequence length
        """
        # if max_seq_len == None meaning there is no token length yet
        if self.max_seq_len and token_length > self.max_seq_len:
            raise HTTPException(status_code=400, detail=f"Token length exceeds max length of {self.max_seq_len}")


    async def chat_no_tool(
        self,
        query: ChatMLQuery,
        prompt_eng: PromptEngine,
        gen_queue, request: Request,
        stop_event: asyncio.Event = None
    ):

        prompt = prompt_eng.get_prompt(
            query,
            thinking_template=query.thinking_template,
            backend=self.backend
        )

        formatter_prefix_regex = self.formatter.regex(
            query.regex_prefix_pattern,
            backend=self.backend,
            preference=query.guided_decoding_backend
        ) if query.regex_prefix_pattern else None

        formatter_regex = self.formatter.regex(
            query.regex_pattern,
            backend=self.backend,
            preference=query.guided_decoding_backend
        ) if query.regex_pattern else None

        token_length_prompt = get_token_length(self.tokenizer, prompt)
        self.validate_token_length(token_length_prompt)

        # think template prompting
        if query.thinking_template:
            try:
                thinking = Thinking(query.thinking_template)
            except Exception as e:
                raise HTTPException(status_code=400, detail=f"thinking_template is not valid XML string")

            # Not returning thinking
            thinking_queue = GenQueue()
            if query.return_thinking:
                # return thinking to front end
                queue_group = [
                    QueueContext.create(gen_queue=thinking_queue, include_GenEnd=True, include_GenStats=False),
                    QueueContext.create(gen_queue=gen_queue, include_GenEnd=False, include_GenStats=False)
                ]
            else:
                # not return thinking to front end
                queue_group = [
                    QueueContext.create(gen_queue=thinking_queue, include_GenEnd=True, include_GenStats=False),
                ]

            await self.generate(
                prompt,
                messages=query.messages,
                gen_type="thinking",
                gen_queue=queue_group,
                temperature=query.temperature,
                top_p=query.top_p,
                prefix_strings=f"<{thinking.root_tag}>",
                stop_words=thinking.root_key_stop_words,
                request=request,
                stop_event=stop_event,
                video=query.video,
            )
            thinking_response, _ = await get_response_from_queue(thinking_queue)

            # Get the new prompt with thinking response
            prompt = prompt_eng.get_prompt(
                query,
                thinking_template=query.thinking_template,
                thinking_response=thinking_response,
                backend=self.backend
            )

        # 1st response if there is regex to match the regex pattern
        first_response = ""

        if query.regex_prefix_pattern:
            prefix_queue = GenQueue()
            queue_group = [
                QueueContext.create(gen_queue=prefix_queue, include_GenEnd=True, include_GenStats=False),
                QueueContext.create(gen_queue=gen_queue, include_GenEnd=False, include_GenStats=False)
            ]

            await self.generate(
                prompt,
                messages=query.messages,
                gen_queue=queue_group,
                temperature=query.temperature,
                top_p=query.top_p,
                formatter=formatter_prefix_regex,
                prefix_strings=query.prefix_strings,
                request=request,
                # stop_words=query.stop_words,
                stop_event=stop_event,
                video=query.video,
            )

            first_response, _ = await get_response_from_queue(prefix_queue)

            # append generated content to the full prompt
            prompt = prompt.strip() + first_response.strip()

        # Final generation to return to client
        # set prefix string
        prefix_strings = None if query.regex_prefix_pattern else query.prefix_strings

        # Handle Artifact mode: overwrite prefix_strings if in artifact mode
        stop_words_to_use = query.stop_words
        banned_strings = None
        if query.artifact and query.artifact == "Fast":
            prefix_strings = None
            manual_prefix_string = "<answer>\n "
            prompt += manual_prefix_string

            # ban XML comment format which could mess up the parsing of output
            banned_strings = ["<![CDATA[", "<!--"]

            # add the stopword for artifact tag to the answer
            if isinstance(stop_words_to_use, list):
                stop_words_to_use.append("</answer>")
            elif isinstance(stop_words_to_use, str):
                stop_words_to_use = [stop_words_to_use, "</answer>"]
            else:
                stop_words_to_use = "</answer>"

            # add the initial string as prefix_strings can not be used together with banned_strings
            chunk = GenText(content=manual_prefix_string)
            gen_queue.put_nowait(chunk)

        await self.generate(
            prompt=prompt,
            messages=query.messages,
            gen_queue=gen_queue,
            **{
                'temperature': query.temperature,
                'top_p': query.top_p,
                'formatter': formatter_regex,
                'stop_words': stop_words_to_use,
                'max_tokens': query.max_tokens,
                'prefix_strings': prefix_strings,  # already generated as part of the prefix string
                'banned_strings': banned_strings,
                'request': request,
                'stop_event': stop_event,
                'video': query.video,
            }
        )

    async def chat_with_tool(
        self,
        query: ChatMLQuery,
        prompt_eng,
        gen_queue,
        request: Request,
        stop_event: asyncio.Event = None
    ):
        # use_tool marker
        use_tool_bool = False  # this will be set to True if tool is used
        fall_back_bool = False  # this will decide if fallback generation with regex enforcement is required

        # tool class have the method to handle converting processing or tool requirement, schema and response
        tool_handler = Tools(
            prompt_eng=prompt_eng,
            tools=query.tools,
            tool_choice=query.tool_choice,
        )

        # substitute the actual list of tool to the thinking template
        tool_thinking_queue = GenQueue()

        tool_thinking_to_use = self.TOOL_THINKING
        if query.tool_choice != "auto":
            tool_thinking_to_use = self.TOOL_FORCE_THINKING

        tool_thinking_formatted = tool_thinking_to_use.xml.format_map(
            {"fstring_available_tools": tool_handler.tool_name_list}
        )

        # perform generation with tool thinking to evaluate if it is necessity to call a tool
        prompt = prompt_eng.get_prompt(
            query,
            pydantic_tool_dict=tool_handler.tool_dict,
            thinking_template=tool_thinking_formatted,
            answer_format_schema=False,
            backend=self.backend
            # leading_prompt=leading_prompt,
        )

        await self.generate(
            prompt,
            messages=query.messages,
            gen_queue=tool_thinking_queue,
            temperature=query.temperature,
            top_p=query.top_p,
            stop_words=tool_thinking_to_use.root_key_stop_words,
            prefix_strings=f"<{tool_thinking_to_use.root_tag}>",
            request=request,
            # formatter=formatter_regex  # no longer enforce format
            stop_event=stop_event,
            video=query.video,
        )

        # evaluate tool usage necessity
        tool_thinking_response, _ = await get_response_from_queue(tool_thinking_queue)

        # see if llm able to generate the xml format correctly
        if query.tool_choice != "auto":
            use_tool_bool = True
        else:
            try:
                # parse the xml object
                tool_thinking_response_dict = Thinking.parse_xml_to_dict(tool_thinking_response)
                tool_decision = tool_thinking_response_dict[tool_thinking_to_use.root_tag]["final_decision"][
                    "is_tool_needed"]  # TODO implement a less hardcoding way

                if tool_decision.lower().strip() == "yes":
                    use_tool_bool = True
                elif tool_decision.lower().strip() == "no":
                    use_tool_bool = False
                else:
                    # the format was not enforce, to perform fall back check
                    fall_back_bool = True

            except Exception as e:
                logger.error(f"XML parsing failed: {e}")
                # Fallback: check for the presence of Yes/No in the raw XML
                yes_pattern = r'<is_tool_needed>\s*Yes\s*</is_tool_needed>'
                no_pattern = r'<is_tool_needed>\s*No\s*</is_tool_needed>'

                if re.search(yes_pattern, tool_thinking_response, re.IGNORECASE):
                    use_tool_bool = True
                elif re.search(no_pattern, tool_thinking_response, re.IGNORECASE):
                    use_tool_bool = False
                else:
                    fall_back_bool = True

        logger.info(tool_thinking_response)

        # Fall back plan
        fall_back_prompt = ""
        if fall_back_bool:
            logger.info("Tool Analysis fallback")
            # generate fall back response with regex enforcement:
            fall_back_prompt = ("\n Fill in the blank in below sentence with either 'needed' or 'not needed'"
                                "In summary, tool calling is {blank}\n"
                                "Answer: blank=")
            prompt += tool_thinking_response + fall_back_prompt

            # perform generation with tool thinking to evaluate if it is necessity
            tool_thinking_queue_fallback = GenQueue()

            formatter_regex = self.formatter.regex('(needed|not needed)', backend=self.backend, preference=query.guided_decoding_backend)

            await self.generate(
                prompt,
                messages=query.messages,
                gen_queue=tool_thinking_queue_fallback,
                temperature=query.temperature,
                top_p=query.top_p,
                request=request,
                # prefix_strings="n",
                # stop_words=TOOL_THINKING.root_key_stop_words,
                formatter=formatter_regex,  # no longer enforce format
                stop_event = stop_event,
                video=query.video,
            )

            # evaluate tool usage necessity
            tool_thinking_decision_fallback, _ = await get_response_from_queue(tool_thinking_queue_fallback)

            # decide if tool call is required
            if "not needed" in tool_thinking_decision_fallback.lower():
                use_tool_bool = False
            elif "needed" in tool_thinking_decision_fallback.lower():
                use_tool_bool = True

        # USE TOOL
        if use_tool_bool:
            # create the pydantic schema to enforce generation
            tool_combined_pydantic_lmfe = create_function_models_v2(tool_handler.tool_dict)

            class ToolCalling_LMFE(ClassSchema):
                """ The format to call one or multiple tools """
                functions_calling: List[Union[tuple(tool_combined_pydantic_lmfe)]] = []

            # create the pydantic schema to enforce generation for formatron which use ClassSchema
            tool_combined_pydantic_formatron = create_function_models_formatron(tool_handler.tool_dict_formatron)
            class ToolCalling_formatron(ClassSchema):
                """ The format to call one or multiple tools """
                functions_calling: List[Union[tuple(tool_combined_pydantic_formatron)]] = []

            formatter_json = self.formatter.json(
                pydantic_model_lmfe=ToolCalling_LMFE,
                pydantic_model_formatron=ToolCalling_formatron,
                backend=self.backend,
                preference = query.guided_decoding_backend
            )

            # Experiment feature, formulate function calling as python programming. Which is more natural than a random Json output as part of conversation
            tool_as_code_prompt = """
def run_function(arg_dict):
    function_calls = arg_dict["functions_calling"]

    if function_calls == []:
        print("No function/tool calling needed")
        return

    for call in function_calls:
        function_name = call["name"]
        arguments = call["arguments"]
        globals()[function_name](**arguments)

# Perform function calling if need to
arg_dict = """

            # get final prompt
            prompt = prompt_eng.get_prompt(
                query,
                thinking_template=self.TOOL_THINKING.xml,
                thinking_response=tool_thinking_response,
                pydantic_tool_dict=tool_handler.tool_dict,
                answer_format_schema=True,
                leading_prompt=(
                    f"{fall_back_prompt}\n"
                    'Now i will convert my answer above into "functions_calling" format by continuing this continue this code.\n'
                    f"{tool_as_code_prompt}"
                ),
                backend=self.backend,
            )
            logger.info(prompt)

            # generate
            await self.generate(
                prompt,
                messages=query.messages,
                gen_queue=gen_queue,
                gen_type=GenStart(gen_type="tool"),
                temperature=query.temperature,
                top_p=query.top_p,
                # stop_words=TOOL_THINKING.root_key_stop_words,
                prefix_strings=['{\n "functions_calling": ['],
                formatter=formatter_json,
                max_tokens=query.max_tokens,
                request=request,
                stop_event=stop_event,
                video=query.video,
            )

        # NOT USE TOOL
        if not use_tool_bool:
            prompt = prompt_eng.get_prompt(
                query,
                backend=self.backend,
            )

            if query.tool_choice == "auto":
                # Normal generation
                await self.generate(
                    prompt,
                    messages=query.messages,
                    gen_queue=gen_queue,
                    gen_type=GenStart(gen_type="text"),
                    temperature=query.temperature,
                    prefix_strings=query.prefix_strings,
                    max_tokens=query.max_tokens,
                    request=request,
                    stop_event=stop_event,
                    video=query.video,
                )
            else:
                # tool choice is forced -> return empty tool calling
                gen_queue.put_nowait(GenStart(gen_type="tool"))
                gen_queue.put_nowait(GenText(content='{"functions_calling":[]}'))
                gen_queue.put_nowait(GenerationStats())
                gen_queue.put_nowait(GenEnd())


    async def _tool_calling_task(
        self,
        prompt_eng: PromptEngine,
        query:ChatMLQuery,
        request: Request,
        tool_handler: Tools,
        tool_as_code_prompt: str
    ) -> ToolCallV2:

        # queue for generating output
        gen_queue_dynamic = [GenQueueDynamic()]

        # create the pydantic schema to enforce generation
        tool_combined_pydantic_lmfe = create_function_models_v2(tool_handler.tool_dict)

        class ToolCalling_LMFE(ClassSchema):
            """ The format to call one or multiple tools """
            functions_calling: List[Union[tuple(tool_combined_pydantic_lmfe)]] = []

        # create the pydantic schema to enforce generation for formatron which use ClassSchema
        tool_combined_pydantic_formatron = create_function_models_formatron(
            tool_handler.tool_dict_formatron)

        class ToolCalling_formatron(ClassSchema):
            """ The format to call one or multiple tools """
            functions_calling: List[Union[tuple(tool_combined_pydantic_formatron)]] = []

        formatter_json = self.formatter.json(
            pydantic_model_lmfe=ToolCalling_LMFE,
            pydantic_model_formatron=ToolCalling_formatron,
            backend=self.backend,
            preference=query.guided_decoding_backend
        )

        # shar prompt with tool decision for better KV
        tool_answer_as_code = """
    
# Perform function calling if need to by continuing this code. Return {"functions_calling": []} if function calling is not needed.
response(to="function", arg_dict="""

        if query.tool_instruction_position == "prefix":
            _prefix_prompt = tool_as_code_prompt
            _leading_prompt = tool_answer_as_code
        else:
            _prefix_prompt = ""
            _leading_prompt = tool_as_code_prompt + tool_answer_as_code

        prompt = prompt_eng.get_prompt(
            query,
            pydantic_tool_dict=tool_handler.tool_dict,
            # pydantic_tool_code=tool_handler.tool_def_as_code,
            answer_format_schema=True,
            backend=self.backend,
            prefix_prompt=_prefix_prompt,
            leading_prompt=_leading_prompt,
        )

        stop_event = asyncio.Event()

        generate_kwargs = {
            "prompt": prompt,
            "messages": query.messages,
            "gen_queue": gen_queue_dynamic,
            "gen_type": GenStart(gen_type="tool"),
            "temperature": query.temperature,
            "top_p": query.top_p,
            "prefix_strings": ['{\n "functions_calling": ['],
            "formatter": formatter_json,
            "max_tokens": query.max_tokens,
            "request": request,
            "stop_event": stop_event,
            "video": query.video
        }

        return ToolCallV2(
            gen_dynamic_queue=gen_queue_dynamic,
            stop_event=stop_event,
            generate_kwargs=generate_kwargs
        )

    async def _tool_decision_task(
        self,
        prompt_eng: PromptEngine,
        query: ChatMLQuery,
        request: Request,
        tool_handler: Tools,
        tool_as_code_prompt: str
    ) -> Tuple[ToolCallV2, Callable[[str], bool]]:
        # queue for generating output
        gen_queue_dynamic = [GenQueueDynamic()]

        # create prompt
        # reuse tool call to share the kv cache as possible
        _tool_answer_prefix = "to="
        _tool_thinking_length_guide = ""
        if query.tool_call_thinking:
            _tool_answer_prefix = "thinking="
            _tool_thinking_length_guide = "# briefly think if tool thinking is needed. Never redo a tool with same arguments."

        tool_decision_answer_as_code_prompt = f"""
{_tool_thinking_length_guide}
response(""" + _tool_answer_prefix

        def tool_decision_check_fn(tool_decision_answer: str) -> bool:
            """ return True if tool usage is true"""
            if tool_decision_answer.lower().endswith('"function"') or tool_decision_answer.lower().endswith("'function'"):
                logger.info(f"Tool decision: {tool_decision_answer}")
                return True
            else:
                return False

        def tool_decision_check_with_thinking_fn(tool_decision_answer: str) -> bool:
            """ return True if tool usage is true"""
            if tool_decision_answer.lower().endswith('"function"') or tool_decision_answer.lower().endswith("'function'"):
                logger.info(f"Tool decision: internal_thinking: {tool_decision_answer}")
                return True
            else:
                return False

        if query.tool_instruction_position=="prefix":
            _prefix_prompt = tool_as_code_prompt
            _postfix_prompt = tool_decision_answer_as_code_prompt
        else:
            _prefix_prompt = ""
            _postfix_prompt = tool_as_code_prompt + tool_decision_answer_as_code_prompt

        prompt = prompt_eng.get_prompt(
            query,
            pydantic_tool_dict=tool_handler.tool_dict,
            # pydantic_tool_code=tool_handler.tool_def_as_code,
            answer_format_schema=True,
            backend=self.backend,
            prefix_prompt=_prefix_prompt,
            leading_prompt=_postfix_prompt
        )

        formatter_regex = self.formatter.regex(
            regex_pattern=r'"user"|"function"',
            backend=self.backend,
            preference=query.guided_decoding_backend
        )

        stop_event = asyncio.Event()

        generate_kwargs = {
            "prompt": prompt,
            "messages": query.messages,
            "gen_queue": gen_queue_dynamic,
            "gen_type": GenStart(gen_type="tool"),
            "temperature": query.temperature,
            "top_p": query.top_p,
            "prefix_strings": '"',
            "formatter": formatter_regex if not query.tool_call_thinking else None,
            "max_tokens": query.tool_call_thinking_token,
            "request": request,
            "stop_event": stop_event,
            "stop_words": ['"user"', '"function"', "'user'", "'function'"],
            "video": query.video
        }

        tool_decision_check_fn_to_use = tool_decision_check_fn if query.tool_call_thinking else tool_decision_check_with_thinking_fn

        return ToolCallV2(
            gen_dynamic_queue=gen_queue_dynamic,
            stop_event=stop_event,
            generate_kwargs=generate_kwargs
        ), tool_decision_check_fn_to_use


    async def _no_tool_task(
        self,
        prompt_eng: PromptEngine,
        query:ChatMLQuery,
        request: Request
    ) -> ToolCallV2:

        gen_queue_dynamic = [GenQueueDynamic()]

        prompt = prompt_eng.get_prompt(
            query,
            backend=self.backend,
        )

        stop_event = asyncio.Event()

        generate_kwargs = {
            "prompt": prompt,
            "messages": query.messages,
            "gen_queue": gen_queue_dynamic,
            "gen_type": GenStart(gen_type="text"),
            "temperature": query.temperature,
            "top_p": query.top_p,
            "prefix_strings": query.prefix_strings,
            "max_tokens": query.max_tokens,
            "request": request,
            "stop_event": stop_event,
            "video": query.video,
        }

        return ToolCallV2(
            gen_dynamic_queue=gen_queue_dynamic,
            stop_event=stop_event,
            generate_kwargs=generate_kwargs
        )



    async def chat_with_tool_v2(
        self,
        query: ChatMLQuery,
        prompt_eng: PromptEngine,
        gen_queue: GenQueueDynamic,
        request: Request,
        stop_event: asyncio.Event = None
    ):
        """
        handle tool calling llm generation
        main purpose is to handle auto model of generation by generate both tool and non tool usage at the same time
        dynamically swap to the best answer by LLM
        This will reduce the wait time from generating sequentially
        """

        # tool class have the method to handle converting processing or tool requirement, schema and response
        tool_handler = Tools(
            prompt_eng=prompt_eng,
            tools=query.tools,
            tool_choice=query.tool_choice,
        )

        # create prompt
        _tool_thinking_fn_header = ""
        _tool_answer_prefix = "to="
        if query.tool_call_thinking:
            _tool_thinking_fn_header = 'internal_thinking: str="",'
            _tool_answer_prefix = "thinking="

        tool_as_code_prompt = """
Function/ Tool calling Instruction:
# Use Function calling if:
- It is needed to answer user question.
- Be conservative and only use function calling if it is necessary and suitable.

# Reply directly to user if:
- It is uncertain if user wants function calling.
- Function calling is not related to user's question.
- Clarification is needed. 
- Call a tool only when needed, and never re-do a tool call that you previously did with the exact same parameters.

When a tool/ function is requested, it's result will be run by the system and be returned with reference to the tool call if available
e.g. Result of tool call reference id <tool_call_id>.
```python
def run_function(arg_dict):
    function_calls = arg_dict["functions_calling"]

    if function_calls == []:
        print("No function/tool calling needed")
        return

    for call in function_calls:
        function_name = call["name"]
        arguments = call["arguments"]
        globals()[function_name](**arguments)
        
allow_functions = [""" + tool_handler.tool_name_list + """]

def response(""" + _tool_thinking_fn_header + """to: Literal["user","function"], arg_dict: Optional[List[]]=[]):
   if to=="user":
       # NO function calling needed
       break  # exit for a normal answer
   elif to=="function" and arg_dict!=[]:
       run_function(arg_dict)
   else:
      raise Exception("Either response to user without function calling or function calling is required.")
```
Example:
# No function calling needed:
response(thinking="because... so function calling is not required ", to="user")
my answer is...

# Function calling needed:
response(thinking="because... so function calling is required", to="function", arg_dict={
  "functions_calling": [{
      "name": "function_name",
      "arguments": {"argument_name": "value"}
    }
]}) 
# answer to user is prohibited if using function calling

# Example of follow-up answer after function calling result provided by user
User: what is 2^3?
Assistant:                                                                                                                                                             
power_number_tool({number: 2, power: 3})
8
Assistant: 2^3 is 8
---
IMPORTANT: 'Request for tool call with reference id' and 'Result of tool call reference id' are auto populated by the system for reference
            DO NOT output Request for tool call with reference id by yourself
End of Function Calling Instruction
---
"""


        tool_call = await self._tool_calling_task(
            prompt_eng=prompt_eng,
            query=query,
            request=request,
            tool_handler=tool_handler,
            tool_as_code_prompt=tool_as_code_prompt
        )

        tool_decision, tool_decision_check_fn = await self._tool_decision_task(
            prompt_eng=prompt_eng,
            query=query,
            request=request,
            tool_handler=tool_handler,
            tool_as_code_prompt=tool_as_code_prompt
        )


        no_tool = await self._no_tool_task(
            prompt_eng=prompt_eng,
            query=query,
            request=request
        )

        tasks = []  # to track the task running

        tool_decision_task = None
        tool_task = None
        no_tool_task = None


        # ensure result is out for decision
        tool_decision_outcome: Literal["user", "function"] = "function"

        if self.support_concurrency:
            # tool task will always be run. It is assumed that if this function is called, tool generation is required
            tool_task = asyncio.create_task(self.generate(**tool_call.generate_kwargs))
            tasks.append(tool_task)

            # if mode is auto, run the tool usage decision generation
            if query.tool_choice == "auto":
                tool_decision_task = asyncio.create_task(self.generate(**tool_decision.generate_kwargs))
                tasks.append(tool_decision_task)

                no_tool_task = asyncio.create_task(self.generate(**no_tool.generate_kwargs))
                tasks.append(no_tool_task)

                tool_decision_outcome, _ = await get_response_from_queue(tool_decision.gen_dynamic_queue)


            if tool_decision_check_fn(tool_decision_outcome):
                logger.info("Tool auto decision: function")

                _has_tool = True

                # check first if tool managed to generate
                if tool_call.gen_dynamic_queue[0].qsize() < 3:
                    await asyncio.sleep(0.3)    # wait for 0.3 more second

                    if tool_call.gen_dynamic_queue[0].qsize() < 3:
                        _has_tool = False

                if _has_tool:
                    if no_tool:
                        no_tool.stop_event.set()

                    # swap queue for output to function calling
                    gen_queue.swap(tool_call.gen_dynamic_queue[0])
                else:
                    # tool not managed to generate, fall back to standard reply
                    tool_call.stop_event.set()

                    # swap queue non function calling
                    gen_queue.swap(no_tool.gen_dynamic_queue[0])

            else:
                logger.info("Tool auto decision: user")
                if tool_call:
                    tool_call.stop_event.set()

                # swap queue non function calling
                gen_queue.swap(no_tool.gen_dynamic_queue[0])

            # Wait for the remaining tasks to complete (if needed)
            try:
                await asyncio.gather(*tasks)
            except Exception as e:
                logger.error(f"Error in one of the tasks: {e}")

        else:   # non concurrency backend
            # first generate what is the decision regarding whether use tool or not
            tool_decision_task = await self.generate(**tool_decision.generate_kwargs)
            tool_decision_outcome, _ = await get_response_from_queue(tool_decision.gen_dynamic_queue)

            # handle tool/ non tool generation
            if tool_decision_check_fn(tool_decision_outcome):
                logger.info("Tool auto decision: function")
                tool_task = asyncio.create_task(self.generate(**tool_call.generate_kwargs))

                # swap queue for output to function calling
                gen_queue.swap(tool_call.gen_dynamic_queue[0])
            else:
                logger.info("Tool auto decision: user")
                no_tool_task = asyncio.create_task(self.generate(**no_tool.generate_kwargs))

                # swap queue non function calling
                gen_queue.swap(no_tool.gen_dynamic_queue[0])




